"""
Author: Gabriele Pompa (gabriele.pompa@gmail.com)

Date: 21-May-2020
File name: numerical_routines.py
"""

# ----------------------- sub-modules imports ------------------------------- #

from .utils import *


# -----------------------------------------------------------------------------#

class NumericGreeks:
    """
    NumericGreeks class: a class implementing finite-differences numerical methods
    to compute main greeks for option contracts.

    Attributes:
    -----------
        FinancialObject (EuropeanOption sub-class or Portfolio):      Instance of an EuropeanOption sub-class 
                                                                      (PlainVanillaOption or DigitalOption) or Portfolio
                                                                      class.
       epsilon (float):                                               tolerance for finite-differences                                                                      

    Public Methods:
    --------   
    
        getter and setter for epsilon attribute
        
        delta: float
            Computes the numeric delta of the FinancialObject.

        theta: float
            Computes the numeric theta of the FinancialObject.

        gamma: float
            Computes the numeric gamma of the FinancialObject.

        vega: float
            Computes the numeric vega of the FinancialObject.

        rho: float
            Computes the numeric rho of the FinancialObject.
        
    Instantiation and Usage examples: 
    --------   
        
        - options_numeric_greeks.py
        - options_numeric_analytic_greeks_comparison.py
    """

    def __init__(self, FinancialObject, epsilon=1e-4):
        self.opt = FinancialObject
        self.f = FinancialObject.price
        self.__eps = epsilon

    def get_epsilon(self):
        return self.__eps

    def set_epsilon(self, eps=1e-4):
        self.__eps = eps

    def delta(self, **kwargs):
        """
        Numeric derivative df/dS.
        
        Can be called as self.opt.price() method.
        """
        # process parameter
        S0 = kwargs["S"] if "S" in kwargs else self.opt.get_S()
        S0 = homogenize(S0)

        # delete "S" from kwargs if it exists
        kwargs.pop("S", None)

        return (self.f(S=S0 + self.get_epsilon(), **kwargs) - self.f(S=S0 - self.get_epsilon(), **kwargs)) / (
                    2 * self.get_epsilon())

    def gamma(self, **kwargs):
        """
        Numeric derivative d^2f/dS^2.
        
        Can be called as self.opt.price() method.
        """
        # process parameter
        S0 = kwargs["S"] if "S" in kwargs else self.opt.get_S()
        S0 = homogenize(S0)

        # delete "S" from kwargs if it exists
        kwargs.pop("S", None)

        return (self.f(S=S0 - self.get_epsilon(), **kwargs) - 2.0 * self.f(S=S0, **kwargs) + self.f(
            S=S0 + self.get_epsilon(), **kwargs)) / (self.get_epsilon() * self.get_epsilon())

    def vega(self, **kwargs):
        """
        Numeric derivative df/dsigma. By default, it is scaled to consider 
        variation of +1% of sigma (not +100%).
        
        Can be called as self.opt.price() method.
        """
        # process parameter
        sigma0 = kwargs["sigma"] if "sigma" in kwargs else self.opt.get_sigma()
        sigma0 = homogenize(sigma0)

        # delete "sigma" from kwargs if it exists
        kwargs.pop("sigma", None)

        # rescaling factor
        rescaling_factor = kwargs["factor"] if "factor" in kwargs else 0.01

        return (self.f(sigma=sigma0 + self.get_epsilon(), **kwargs) - self.f(sigma=sigma0 - self.get_epsilon(),
                                                                             **kwargs)) / (
                           2 * self.get_epsilon()) * rescaling_factor

    def theta(self, **kwargs):
        """
        Numeric derivative df/dt = -df/dtau. By default, it is scaled to consider 
        variation of +1 calendar day of t (not +1 year).
        
        Can be called as self.opt.price() method.
        """
        # process parameter
        tau0 = kwargs["tau"] if "tau" in kwargs else self.opt.get_tau()
        tau0 = homogenize(tau0, reverse_order=True)

        # delete "tau" from kwargs if it exists
        kwargs.pop("tau", None)

        # rescaling factor
        rescaling_factor = kwargs["factor"] if "factor" in kwargs else 1.0 / 365.0

        return -((self.f(tau=tau0 + self.get_epsilon(), **kwargs) - self.f(tau=tau0 - self.get_epsilon(), **kwargs)) / (
                    2 * self.get_epsilon())) * rescaling_factor

    def rho(self, **kwargs):
        """
        Numeric derivative df/dr. By default, it is scaled to consider 
        variation of +1% of r (not +100%).
        
        Can be called as self.opt.price() method.
        """
        # process parameter
        r0 = kwargs["r"] if "r" in kwargs else self.opt.get_r()
        r0 = homogenize(r0)

        # delete "r" from kwargs if it exists
        kwargs.pop("r", None)

        # rescaling factor
        rescaling_factor = kwargs["factor"] if "factor" in kwargs else 0.01

        return ((self.f(r=r0 + self.get_epsilon(), **kwargs) - self.f(r=r0 - self.get_epsilon(), **kwargs)) / (
                    2 * self.get_epsilon())) * rescaling_factor



"""
Created by: Gabriele Pompa (gabriele.pompa@gmail.com)

File: market.py

Created on Tue Jul 14 2020 - Version: 1.0

Description: 
    
This file contains the definition of MarketEnvironment class.
"""

# ----------------------- sub-modules imports ------------------------------- #

from ..utils.utils import date_string_to_datetime_obj


# -----------------------------------------------------------------------------#

class MarketEnvironment:
    """
    MarketEnvironment class: setting the market environment: date, underlying level, volatility level and short-rate.
    
    Attributes:
    -----------
        t (str; dt.datetime):     valuation date. Can be either a "dd-mm-YYYY" String or a dt.datetime object
        r (float):                continuously compounded short-rate;
        S_t (float):              spot price of the underlying asset at the valuation date 't';
        sigma (float):            volatility of underlying asset;
    
    Public Methods:
    --------   
    
        getters and setters for all attributes
        
    Usage examples: 
    --------   
    
        See docstrings for EuropeanOption and Portfolio classes.

    Instantiation: 
    --------   
        - default: MarketEnvironment() is equivalent to 
                   MarketEnvironment(t="19-04-2020", r=0.05, S_t=90.0, sigma=0.2)

        - general: MarketEnvironment(t="DD-MM-YYYY" String, r=Float, S_t=Float, sigma=Float)
    """

    def __init__(self, t="19-04-2020", r=0.05, S_t=90.0, sigma=0.2):
        print("Initializing the MarketEnvironment!")

        self.__t = date_string_to_datetime_obj(t)
        self.__r = r
        self.__S = S_t
        self.__sigma = sigma

    def __repr__(self):
        return r"MarketEnvironment(t={}, r={:.1f}%, S_t={:.1f}, sigma={:.1f}%)". \
            format(self.get_t().strftime("%d-%m-%Y"), self.get_r() * 100, self.get_S(), self.get_sigma() * 100)

    # getters
    def get_t(self):
        return self.__t

    def get_r(self):
        return self.__r

    def get_S(self):
        return self.__S

    def get_sigma(self):
        return self.__sigma

    # setters 
    def set_t(self, t):
        self.__t = date_string_to_datetime_obj(t)

    def set_r(self, r):
        self.__r = r

    def set_S(self, S):
        self.__S = S

    def set_sigma(self, sigma):
        self.__sigma = sigma



import unittest
import copy
import numpy as np
import numpy.testing as np_test
import pandas as pd
import pandas.testing as pd_test
import warnings

from pyblackscholesanalytics.market.market import MarketEnvironment
from pyblackscholesanalytics.options.options import PlainVanillaOption, DigitalOption
from pyblackscholesanalytics.utils.utils import scalarize


class TestPlainVanillaOption(unittest.TestCase):
    """Class to test public methods of PlainVanillaOption class"""

    def setUp(self) -> None:
        warnings.filterwarnings("ignore")

        # common market environment
        mkt_env = MarketEnvironment()

        # option objects
        self.call_opt = PlainVanillaOption(mkt_env)
        self.put_opt = PlainVanillaOption(mkt_env, option_type="put")

        # pricing parameters
        S_scalar = 100
        S_vector = [90, 100, 110]
        t_scalar_string = "01-06-2020"
        t_date_range = pd.date_range(start="2020-04-19", end="2020-12-21", periods=5)

        # common pricing parameter setup
        common_params = {"np_output": True, "minimization_method": "Least-Squares"}

        # scalar parameters setup
        self.scalar_params = copy.deepcopy(common_params)
        self.scalar_params["S"] = S_scalar
        self.scalar_params["t"] = t_scalar_string

        # vector parameters setup
        self.vector_params = copy.deepcopy(common_params)
        self.vector_params["S"] = S_vector
        self.vector_params["t"] = t_date_range

        # complex pricing parameter setup
        # (S scalar, K and t vector, sigma distributed as Kxt grid, r distributed as Kxt grid)
        K_vector = [75, 85, 90, 95]
        mK = len(K_vector)
        n = 3
        sigma_grid_K = np.array([0.1 * (1 + i) for i in range(mK * n)]).reshape(n, mK)
        r_grid_K = np.array([0.01 * (1 + i) for i in range(mK * n)]).reshape(n, mK)
        self.complex_params = {"S": S_vector[0],
                               "K": K_vector,
                               "t": pd.date_range(start="2020-04-19", end="2020-12-21", periods=n),
                               "sigma": sigma_grid_K,
                               "r": r_grid_K,
                               "np_output": False,
                               "minimization_method": "Least-Squares"}

    def test_price_scalar(self):
        """Test price - scalar case"""

        # call
        test_call = scalarize(self.call_opt.price(**self.scalar_params))
        expected_call = 7.548381716811839
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.price(**self.scalar_params))
        expected_put = 4.672730506407959
        self.assertEqual(test_put, expected_put)

    def test_price_vector_np(self):
        """Test price - np.ndarray output case"""

        # call
        test_call = self.call_opt.price(**self.vector_params)
        expected_call = np.array([[3.48740247e+00, 8.42523213e+00, 1.55968082e+01],
                                  [2.53045128e+00, 7.14167587e+00, 1.43217796e+01],
                                  [1.56095778e+00, 5.72684668e+00, 1.29736886e+01],
                                  [5.89165298e-01, 4.00605304e+00, 1.14939139e+01],
                                  [7.21585753e-04, 1.38927959e+00, 1.01386434e+01]])
        np_test.assert_allclose(test_call, expected_call)

        # put
        test_put = self.put_opt.price(**self.vector_params)
        expected_put = np.array([[1.00413306e+01, 4.97916024e+00, 2.15073633e+00],
                                 [9.90791873e+00, 4.51914332e+00, 1.69924708e+00],
                                 [9.75553655e+00, 3.92142545e+00, 1.16826738e+00],
                                 [9.62127704e+00, 3.03816479e+00, 5.26025639e-01],
                                 [9.86382907e+00, 1.25238707e+00, 1.75090342e-03]])
        np_test.assert_allclose(test_put, expected_put)

    def test_price_vector_df(self):
        """Test price - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.price(**self.vector_params)
        expected_call = pd.DataFrame(data=[[3.48740247e+00, 8.42523213e+00, 1.55968082e+01],
                                           [2.53045128e+00, 7.14167587e+00, 1.43217796e+01],
                                           [1.56095778e+00, 5.72684668e+00, 1.29736886e+01],
                                           [5.89165298e-01, 4.00605304e+00, 1.14939139e+01],
                                           [7.21585753e-04, 1.38927959e+00, 1.01386434e+01]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call)

        # put
        test_put = self.put_opt.price(**self.vector_params)
        expected_put = pd.DataFrame(data=[[1.00413306e+01, 4.97916024e+00, 2.15073633e+00],
                                          [9.90791873e+00, 4.51914332e+00, 1.69924708e+00],
                                          [9.75553655e+00, 3.92142545e+00, 1.16826738e+00],
                                          [9.62127704e+00, 3.03816479e+00, 5.26025639e-01],
                                          [9.86382907e+00, 1.25238707e+00, 1.75090342e-03]],
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])

        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put)

    def test_PnL_scalar(self):
        """Test P&L - scalar case"""

        # call
        test_call = scalarize(self.call_opt.PnL(**self.scalar_params))
        expected_call = 4.060979245868182
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.PnL(**self.scalar_params))
        expected_put = -5.368600081057167
        self.assertEqual(test_put, expected_put)

    def test_PnL_vector_np(self):
        """Test P&L - np.ndarray output case"""

        # call
        test_call = self.call_opt.PnL(**self.vector_params)
        expected_call = np.array([[0., 4.93782966, 12.10940574],
                                  [-0.95695119, 3.6542734, 10.83437716],
                                  [-1.92644469, 2.2394442, 9.48628613],
                                  [-2.89823717, 0.51865057, 8.00651142],
                                  [-3.48668089, -2.09812288, 6.65124095]])
        np_test.assert_allclose(test_call, expected_call)

        # put
        test_put = self.put_opt.PnL(**self.vector_params)
        expected_put = np.array([[0., -5.06217034, -7.89059426],
                                 [-0.13341186, -5.52218727, -8.3420835],
                                 [-0.28579403, -6.11990513, -8.87306321],
                                 [-0.42005355, -7.0031658, -9.51530495],
                                 [-0.17750152, -8.78894351, -10.03957968]])
        np_test.assert_allclose(test_put, expected_put)

    def test_PnL_vector_df(self):
        """Test P&L - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.PnL(**self.vector_params)
        expected_call = pd.DataFrame(data=[[0., 4.93782966, 12.10940574],
                                           [-0.95695119, 3.6542734, 10.83437716],
                                           [-1.92644469, 2.2394442, 9.48628613],
                                           [-2.89823717, 0.51865057, 8.00651142],
                                           [-3.48668089, -2.09812288, 6.65124095]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call)

        # put
        test_put = self.put_opt.PnL(**self.vector_params)
        expected_put = pd.DataFrame(data=[[0., -5.06217034, -7.89059426],
                                          [-0.13341186, -5.52218727, -8.3420835],
                                          [-0.28579403, -6.11990513, -8.87306321],
                                          [-0.42005355, -7.0031658, -9.51530495],
                                          [-0.17750152, -8.78894351, -10.03957968]],
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])

        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put)

    def test_delta_scalar(self):
        """Test Delta - scalar case"""

        # call
        test_call = scalarize(self.call_opt.delta(**self.scalar_params))
        expected_call = 0.6054075531684143
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.delta(**self.scalar_params))
        expected_put = -0.3945924468315857
        self.assertEqual(test_put, expected_put)

    def test_delta_vector_np(self):
        """Test Delta - np.ndarray output case"""

        # call
        test_call = self.call_opt.delta(**self.vector_params)
        expected_call = np.array([[3.68466757e-01, 6.15283790e-01, 8.05697003e-01],
                                  [3.20097309e-01, 6.00702480e-01, 8.18280131e-01],
                                  [2.54167521e-01, 5.83663527e-01, 8.41522350e-01],
                                  [1.49152172e-01, 5.61339299e-01, 8.91560577e-01],
                                  [8.89758553e-04, 5.23098767e-01, 9.98343116e-01]])
        np_test.assert_allclose(test_call, expected_call)

        # put
        test_put = self.put_opt.delta(**self.vector_params)
        expected_put = np.array([[-0.63153324, -0.38471621, -0.194303],
                                 [-0.67990269, -0.39929752, -0.18171987],
                                 [-0.74583248, -0.41633647, -0.15847765],
                                 [-0.85084783, -0.4386607, -0.10843942],
                                 [-0.99911024, -0.47690123, -0.00165688]])
        np_test.assert_allclose(test_put, expected_put, rtol=5e-6)

    def test_delta_vector_df(self):
        """Test Delta - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.delta(**self.vector_params)
        expected_call = pd.DataFrame(data=[[3.68466757e-01, 6.15283790e-01, 8.05697003e-01],
                                           [3.20097309e-01, 6.00702480e-01, 8.18280131e-01],
                                           [2.54167521e-01, 5.83663527e-01, 8.41522350e-01],
                                           [1.49152172e-01, 5.61339299e-01, 8.91560577e-01],
                                           [8.89758553e-04, 5.23098767e-01, 9.98343116e-01]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call)

        # put
        test_put = self.put_opt.delta(**self.vector_params)
        expected_put = pd.DataFrame(data=[[-0.63153324, -0.38471621, -0.194303],
                                          [-0.67990269, -0.39929752, -0.18171987],
                                          [-0.74583248, -0.41633647, -0.15847765],
                                          [-0.85084783, -0.4386607, -0.10843942],
                                          [-0.99911024, -0.47690123, -0.00165688]],
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])

        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put)

    def test_gamma_scalar(self):
        """Test Gamma - scalar case"""

        # call
        test_call = scalarize(self.call_opt.gamma(**self.scalar_params))
        expected_call = 0.025194958512498786
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.gamma(**self.scalar_params))
        expected_put = copy.deepcopy(expected_call)
        self.assertEqual(test_put, expected_put)

        # assert call and put gamma coincide
        self.assertEqual(test_call, test_put)

    def test_gamma_vector_np(self):
        """Test Gamma - np.ndarray output case"""

        # call
        test_call = self.call_opt.gamma(**self.vector_params)
        expected_call = np.array([[0.02501273, 0.02281654, 0.01493167],
                                  [0.02725456, 0.02648423, 0.01645793],
                                  [0.02950243, 0.03231528, 0.01820714],
                                  [0.02925862, 0.0446913, 0.01918121],
                                  [0.00101516, 0.12030889, 0.00146722]])
        np_test.assert_allclose(test_call, expected_call, rtol=5e-6)

        # put
        test_put = self.put_opt.gamma(**self.vector_params)
        expected_put = copy.deepcopy(expected_call)
        np_test.assert_allclose(test_put, expected_put, rtol=5e-6)

        # assert call and put gamma coincide
        np_test.assert_allclose(test_call, test_put)

    def test_gamma_vector_df(self):
        """Test Gamma - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.gamma(**self.vector_params)
        expected_call = pd.DataFrame(data=[[0.02501273, 0.02281654, 0.01493167],
                                           [0.02725456, 0.02648423, 0.01645793],
                                           [0.02950243, 0.03231528, 0.01820714],
                                           [0.02925862, 0.0446913, 0.01918121],
                                           [0.00101516, 0.12030889, 0.00146722]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call)

        # put
        test_put = self.put_opt.gamma(**self.vector_params)
        expected_put = copy.deepcopy(expected_call)
        pd_test.assert_frame_equal(test_put, expected_put)

        # assert call and put gamma coincide
        pd_test.assert_frame_equal(test_call, test_put)

    def test_vega_scalar(self):
        """Test Vega - scalar case"""

        # call
        test_call = scalarize(self.call_opt.vega(**self.scalar_params))
        expected_call = 0.29405622811847903
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.vega(**self.scalar_params))
        expected_put = copy.deepcopy(expected_call)
        self.assertEqual(test_put, expected_put)

        # assert call and put vega coincide
        self.assertEqual(test_call, test_put)

    def test_vega_vector_np(self):
        """Test Vega - np.ndarray output case"""

        # call
        test_call = self.call_opt.vega(**self.vector_params)
        expected_call = np.array([[0.28419942, 0.32005661, 0.2534375],
                                  [0.23467293, 0.28153094, 0.21168961],
                                  [0.17415326, 0.23550311, 0.16055207],
                                  [0.09220072, 0.17386752, 0.09029355],
                                  [0.00045056, 0.06592268, 0.00097279]])
        np_test.assert_allclose(test_call, expected_call, rtol=1e-5)

        # put
        test_put = self.put_opt.vega(**self.vector_params)
        expected_put = copy.deepcopy(expected_call)
        np_test.assert_allclose(test_put, expected_put, rtol=1e-5)

        # assert call and put vega coincide
        np_test.assert_allclose(test_call, test_put)

    def test_vega_vector_df(self):
        """Test Vega - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.vega(**self.vector_params)
        expected_call = pd.DataFrame(data=[[0.28419942, 0.32005661, 0.2534375],
                                           [0.23467293, 0.28153094, 0.21168961],
                                           [0.17415326, 0.23550311, 0.16055207],
                                           [0.09220072, 0.17386752, 0.09029355],
                                           [0.00045056, 0.06592268, 0.00097279]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call, check_less_precise=True)

        # put
        test_put = self.put_opt.vega(**self.vector_params)
        expected_put = copy.deepcopy(expected_call)
        pd_test.assert_frame_equal(test_put, expected_put, check_less_precise=True)

        # assert call and put vega coincide
        pd_test.assert_frame_equal(test_call, test_put)

    def test_theta_scalar(self):
        """Test Theta - scalar case"""

        # call
        test_call = scalarize(self.call_opt.theta(**self.scalar_params))
        expected_call = -0.021064685979455443
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.theta(**self.scalar_params))
        expected_put = -0.007759980665812141
        self.assertEqual(test_put, expected_put)

    def test_theta_vector_np(self):
        """Test Theta - np.ndarray output case"""

        # call
        test_call = self.call_opt.theta(**self.vector_params)
        expected_call = np.array([[-0.01516655, -0.01977662, -0.01990399],
                                  [-0.01569631, -0.02176239, -0.0212802],
                                  [-0.01601397, -0.02491789, -0.02297484],
                                  [-0.01474417, -0.03162919, -0.02457737],
                                  [-0.00046144, -0.0728981, -0.01462746]])
        np_test.assert_allclose(test_call, expected_call, rtol=5e-4)

        # put
        test_put = self.put_opt.theta(**self.vector_params)
        expected_put = np.array([[-0.00193999, -0.00655005, -0.00667743],
                                 [-0.00235693, -0.00842301, -0.00794082],
                                 [-0.00256266, -0.01146658, -0.00952353],
                                 [-0.00117813, -0.01806315, -0.01101133],
                                 [0.01321844, -0.05921823, -0.00094758]])
        np_test.assert_allclose(test_put, expected_put, rtol=1e-5)

    def test_theta_vector_df(self):
        """Test Theta - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.theta(**self.vector_params)
        expected_call = pd.DataFrame(data=[[-0.01516655, -0.01977662, -0.01990399],
                                           [-0.01569631, -0.02176239, -0.0212802],
                                           [-0.01601397, -0.02491789, -0.02297484],
                                           [-0.01474417, -0.03162919, -0.02457737],
                                           [-0.00046144, -0.0728981, -0.01462746]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call, check_less_precise=True)

        # put
        test_put = self.put_opt.theta(**self.vector_params)
        expected_put = pd.DataFrame(data=[[-0.00193999, -0.00655005, -0.00667743],
                                          [-0.00235693, -0.00842301, -0.00794082],
                                          [-0.00256266, -0.01146658, -0.00952353],
                                          [-0.00117813, -0.01806315, -0.01101133],
                                          [0.01321844, -0.05921823, -0.00094758]],
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])
        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put, check_less_precise=True)

    def test_rho_scalar(self):
        """Test Rho - scalar case"""

        # call
        test_call = scalarize(self.call_opt.rho(**self.scalar_params))
        expected_call = 0.309243166487844
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.rho(**self.scalar_params))
        expected_put = -0.2575372798733608
        self.assertEqual(test_put, expected_put)

    def test_rho_vector_np(self):
        """Test Rho - np.ndarray output case"""

        # call
        test_call = self.call_opt.rho(**self.vector_params)
        expected_call = np.array([[2.08128741e-01, 3.72449469e-01, 5.12209444e-01],
                                  [1.39670999e-01, 2.81318986e-01, 4.02292404e-01],
                                  [7.76651463e-02, 1.91809707e-01, 2.90026614e-01],
                                  [2.49657984e-02, 1.01399432e-01, 1.68411513e-01],
                                  [2.17415573e-05, 1.39508485e-02, 2.73093423e-02]])
        np_test.assert_allclose(test_call, expected_call, rtol=1e-5)

        # put
        test_put = self.put_opt.rho(**self.vector_params)
        expected_put = np.array([[-4.69071412e-01, -3.04750685e-01, -1.64990710e-01],
                                 [-3.77896910e-01, -2.36248923e-01, -1.15275505e-01],
                                 [-2.80139757e-01, -1.65995197e-01, -6.77782897e-02],
                                 [-1.67672008e-01, -9.12383748e-02, -2.42262934e-02],
                                 [-2.73380139e-02, -1.34089069e-02, -5.04131783e-05]])
        np_test.assert_allclose(test_put, expected_put, rtol=1e-5)

    def test_rho_vector_df(self):
        """Test Theta - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.rho(**self.vector_params)
        expected_call = pd.DataFrame(data=[[2.08128741e-01, 3.72449469e-01, 5.12209444e-01],
                                           [1.39670999e-01, 2.81318986e-01, 4.02292404e-01],
                                           [7.76651463e-02, 1.91809707e-01, 2.90026614e-01],
                                           [2.49657984e-02, 1.01399432e-01, 1.68411513e-01],
                                           [2.17415573e-05, 1.39508485e-02, 2.73093423e-02]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call, check_less_precise=True)

        # put
        test_put = self.put_opt.rho(**self.vector_params)
        expected_put = pd.DataFrame(data=[[-4.69071412e-01, -3.04750685e-01, -1.64990710e-01],
                                          [-3.77896910e-01, -2.36248923e-01, -1.15275505e-01],
                                          [-2.80139757e-01, -1.65995197e-01, -6.77782897e-02],
                                          [-1.67672008e-01, -9.12383748e-02, -2.42262934e-02],
                                          [-2.73380139e-02, -1.34089069e-02, -5.04131783e-05]],
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])
        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put, check_less_precise=True)

    def test_Implied_Vol_scalar(self):
        """Test Implied Volatility - scalar case"""

        # call
        test_call = scalarize(self.call_opt.implied_volatility(**self.scalar_params))
        expected_call = 0.2
        self.assertAlmostEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.implied_volatility(**self.scalar_params))
        expected_put = 0.2
        self.assertAlmostEqual(test_put, expected_put)

    def test_Implied_Vol_vector_np(self):
        """Test Implied Volatility - np.ndarray output case"""

        # call
        test_call = self.call_opt.implied_volatility(**self.vector_params)
        expected_call = 0.2 + np.zeros_like(test_call)
        np_test.assert_allclose(test_call, expected_call, rtol=1e-5)

        # put
        test_put = self.put_opt.implied_volatility(**self.vector_params)
        expected_put = 0.2 + np.zeros_like(test_put)
        np_test.assert_allclose(test_put, expected_put, rtol=1e-5)

    def test_Implied_Vol_vector_df(self):
        """Test Implied Volatility - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.implied_volatility(**self.vector_params)
        expected_call = pd.DataFrame(data=0.2 + np.zeros_like(test_call),
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call, check_less_precise=True)

        # put
        test_put = self.put_opt.implied_volatility(**self.vector_params)
        expected_put = pd.DataFrame(data=0.2 + np.zeros_like(test_put),
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])
        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put, check_less_precise=True)

    def test_complex_parameters_setup(self):
        """
        Test complex parameter setup:
        (S scalar, K and t vector, sigma distributed as Kxt grid, r distributed as Kxt grid)
        """

        # call
        test_call_price = self.call_opt.price(**self.complex_params)
        test_call_PnL = self.call_opt.PnL(**self.complex_params)
        test_call_delta = self.call_opt.delta(**self.complex_params)
        test_call_gamma = self.call_opt.gamma(**self.complex_params)
        test_call_vega = self.call_opt.vega(**self.complex_params)
        test_call_theta = self.call_opt.theta(**self.complex_params)
        test_call_rho = self.call_opt.rho(**self.complex_params)
        test_call_iv = self.call_opt.implied_volatility(**self.complex_params)

        expected_call_price = pd.DataFrame(data=[[15.55231058, 9.40714796, 9.87150919, 10.97983523],
                                                 [20.05777231, 16.15277891, 16.02977848, 16.27588191],
                                                 [15.81433361, 8.75227505, 6.65476799, 5.19785143]],
                                           index=self.complex_params["t"],
                                           columns=self.complex_params["K"])
        expected_call_price.rename_axis("K", axis='columns', inplace=True)
        expected_call_price.rename_axis("t", axis='rows', inplace=True)

        expected_call_PnL = pd.DataFrame(data=[[12.06490811, 5.91974549, 6.38410672, 7.49243276],
                                               [16.57036984, 12.66537644, 12.54237601, 12.78847944],
                                               [12.32693114, 5.26487258, 3.16736552, 1.71044896]],
                                         index=self.complex_params["t"],
                                         columns=self.complex_params["K"])
        expected_call_PnL.rename_axis("K", axis='columns', inplace=True)
        expected_call_PnL.rename_axis("t", axis='rows', inplace=True)

        expected_call_delta = pd.DataFrame(data=[[0.98935079, 0.69453583, 0.58292013, 0.53579465],
                                                 [0.79256302, 0.65515368, 0.60705014, 0.57529078],
                                                 [0.90573251, 0.6717088, 0.54283905, 0.43788167]],
                                           index=self.complex_params["t"],
                                           columns=self.complex_params["K"])
        expected_call_delta.rename_axis("K", axis='columns', inplace=True)
        expected_call_delta.rename_axis("t", axis='rows', inplace=True)

        expected_call_gamma = pd.DataFrame(data=[[0.00373538, 0.02325203, 0.01726052, 0.01317896],
                                                 [0.01053321, 0.01130107, 0.01011038, 0.0090151],
                                                 [0.01253481, 0.0242596, 0.02420515, 0.02204576]],
                                           index=self.complex_params["t"],
                                           columns=self.complex_params["K"])
        expected_call_gamma.rename_axis("K", axis='columns', inplace=True)
        expected_call_gamma.rename_axis("t", axis='rows', inplace=True)

        expected_call_vega = pd.DataFrame(data=[[0.02122104, 0.26419398, 0.29417607, 0.29948378],
                                                [0.15544424, 0.20013116, 0.20888592, 0.2128651],
                                                [0.02503527, 0.05383637, 0.05908709, 0.05870816]],
                                          index=self.complex_params["t"],
                                          columns=self.complex_params["K"])
        expected_call_vega.rename_axis("K", axis='columns', inplace=True)
        expected_call_vega.rename_axis("t", axis='rows', inplace=True)

        expected_call_theta = pd.DataFrame(data=[[-0.00242788, -0.01322973, -0.02073753, -0.02747845],
                                                 [-0.03624253, -0.0521798, -0.06237363, -0.07180046],
                                                 [-0.12885912, -0.28334665, -0.33769702, -0.36349655]],
                                           index=self.complex_params["t"],
                                           columns=self.complex_params["K"])
        expected_call_theta.rename_axis("K", axis='columns', inplace=True)
        expected_call_theta.rename_axis("t", axis='rows', inplace=True)

        expected_call_rho = pd.DataFrame(data=[[0.51543152, 0.37243495, 0.29872256, 0.26120194],
                                               [0.18683002, 0.15599644, 0.14066931, 0.12935721],
                                               [0.01800044, 0.0141648, 0.01156185, 0.00937301]],
                                         index=self.complex_params["t"],
                                         columns=self.complex_params["K"])
        expected_call_rho.rename_axis("K", axis='columns', inplace=True)
        expected_call_rho.rename_axis("t", axis='rows', inplace=True)

        expected_call_iv = pd.DataFrame(data=self.complex_params["sigma"],
                                        index=self.complex_params["t"],
                                        columns=self.complex_params["K"])
        expected_call_iv.rename_axis("K", axis='columns', inplace=True)
        expected_call_iv.rename_axis("t", axis='rows', inplace=True)

        pd_test.assert_frame_equal(test_call_price, expected_call_price)
        pd_test.assert_frame_equal(test_call_PnL, expected_call_PnL)
        pd_test.assert_frame_equal(test_call_delta, expected_call_delta)
        pd_test.assert_frame_equal(test_call_gamma, expected_call_gamma)
        pd_test.assert_frame_equal(test_call_vega, expected_call_vega)
        pd_test.assert_frame_equal(test_call_theta, expected_call_theta)
        pd_test.assert_frame_equal(test_call_rho, expected_call_rho)
        pd_test.assert_frame_equal(test_call_iv, expected_call_iv)

        # put
        test_put_price = self.put_opt.price(**self.complex_params)
        test_put_PnL = self.put_opt.PnL(**self.complex_params)
        test_put_delta = self.put_opt.delta(**self.complex_params)
        test_put_gamma = self.put_opt.gamma(**self.complex_params)
        test_put_vega = self.put_opt.vega(**self.complex_params)
        test_put_theta = self.put_opt.theta(**self.complex_params)
        test_put_rho = self.put_opt.rho(**self.complex_params)
        test_put_iv = self.put_opt.implied_volatility(**self.complex_params)

        expected_put_price = pd.DataFrame(data=[[0.02812357, 3.22314287, 7.9975943, 13.35166847],
                                                [3.70370639, 9.31459014, 13.76319167, 18.54654119],
                                                [0.62962992, 3.51971706, 6.38394341, 9.88603552]],
                                          index=self.complex_params["t"],
                                          columns=self.complex_params["K"])
        expected_put_price.rename_axis("K", axis='columns', inplace=True)
        expected_put_price.rename_axis("t", axis='rows', inplace=True)

        expected_put_PnL = pd.DataFrame(data=[[-10.01320701, -6.81818772, -2.04373628, 3.31033788],
                                              [-6.3376242, -0.72674045, 3.72186108, 8.5052106],
                                              [-9.41170067, -6.52161353, -3.65738717, -0.15529507]],
                                        index=self.complex_params["t"],
                                        columns=self.complex_params["K"])
        expected_put_PnL.rename_axis("K", axis='columns', inplace=True)
        expected_put_PnL.rename_axis("t", axis='rows', inplace=True)

        expected_put_delta = pd.DataFrame(data=[[-0.01064921, -0.30546417, -0.41707987, -0.46420535],
                                                [-0.20743698, -0.34484632, -0.39294986, -0.42470922],
                                                [-0.09426749, -0.3282912, -0.45716095, -0.56211833]],
                                          index=self.complex_params["t"],
                                          columns=self.complex_params["K"])
        expected_put_delta.rename_axis("K", axis='columns', inplace=True)
        expected_put_delta.rename_axis("t", axis='rows', inplace=True)

        expected_put_gamma = copy.deepcopy(expected_call_gamma)

        expected_put_vega = copy.deepcopy(expected_call_vega)

        expected_put_theta = pd.DataFrame(data=[[-0.00038744, -0.00863707, -0.01349429, -0.01735551],
                                                [-0.02615404, -0.03850937, -0.04554804, -0.05157676],
                                                [-0.11041151, -0.26012269, -0.31065535, -0.33236619]],
                                          index=self.complex_params["t"],
                                          columns=self.complex_params["K"])
        expected_put_theta.rename_axis("K", axis='columns', inplace=True)
        expected_put_theta.rename_axis("t", axis='rows', inplace=True)

        expected_put_rho = pd.DataFrame(data=[[-0.00691938, -0.21542518, -0.31936724, -0.38666626],
                                              [-0.08152366, -0.14703153, -0.17901683, -0.2068619],
                                              [-0.00249691, -0.00905916, -0.01302149, -0.01656895]],
                                        index=self.complex_params["t"],
                                        columns=self.complex_params["K"])
        expected_put_rho.rename_axis("K", axis='columns', inplace=True)
        expected_put_rho.rename_axis("t", axis='rows', inplace=True)

        expected_put_iv = pd.DataFrame(data=self.complex_params["sigma"],
                                       index=self.complex_params["t"],
                                       columns=self.complex_params["K"])
        expected_put_iv.rename_axis("K", axis='columns', inplace=True)
        expected_put_iv.rename_axis("t", axis='rows', inplace=True)

        pd_test.assert_frame_equal(test_put_price, expected_put_price)
        pd_test.assert_frame_equal(test_put_PnL, expected_put_PnL)
        pd_test.assert_frame_equal(test_put_delta, expected_put_delta)
        pd_test.assert_frame_equal(test_put_gamma, expected_put_gamma)
        pd_test.assert_frame_equal(test_put_vega, expected_put_vega)
        pd_test.assert_frame_equal(test_put_theta, expected_put_theta, check_less_precise=True)
        pd_test.assert_frame_equal(test_put_rho, expected_put_rho)
        pd_test.assert_frame_equal(test_put_iv, expected_put_iv)

        # test gamma and vega consistency
        pd_test.assert_frame_equal(test_call_gamma, test_put_gamma)
        pd_test.assert_frame_equal(test_call_vega, test_put_vega)


class TestDigitalOption(unittest.TestCase):
    """Class to test public methods of DigitalOption class"""

    def setUp(self) -> None:
        warnings.filterwarnings("ignore")

        # common market environment
        mkt_env = MarketEnvironment()

        # option objects
        self.call_opt = DigitalOption(mkt_env)
        self.put_opt = DigitalOption(mkt_env, option_type="put")

        # pricing parameters
        S_scalar = 100
        S_vector = [90, 100, 110]
        t_scalar_string = "01-06-2020"
        t_date_range = pd.date_range(start="2020-04-19", end="2020-12-21", periods=5)

        # common pricing parameter setup
        common_params = {"np_output": True, "minimization_method": "Least-Squares"}

        # scalar parameters setup
        self.scalar_params = copy.deepcopy(common_params)
        self.scalar_params["S"] = S_scalar
        self.scalar_params["t"] = t_scalar_string

        # vector parameters setup
        self.vector_params = copy.deepcopy(common_params)
        self.vector_params["S"] = S_vector
        self.vector_params["t"] = t_date_range

        # complex pricing parameter setup
        # (S scalar, K and t vector, sigma distributed as Kxt grid, r distributed as Kxt grid)
        K_vector = [75, 85, 90, 95]
        mK = len(K_vector)
        n = 3
        sigma_grid_K = np.array([0.1 * (1 + i) for i in range(mK * n)]).reshape(n, mK)
        r_grid_K = np.array([0.01 * (1 + i) for i in range(mK * n)]).reshape(n, mK)
        self.complex_params = {"S": S_vector[0],
                               "K": K_vector,
                               "t": pd.date_range(start="2020-04-19", end="2020-12-21", periods=n),
                               "sigma": sigma_grid_K,
                               "r": r_grid_K,
                               "np_output": False,
                               "minimization_method": "Least-Squares"}

    def test_price_scalar(self):
        """Test price - scalar case"""

        # call
        test_call = scalarize(self.call_opt.price(**self.scalar_params))
        expected_call = 0.529923736000296
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.price(**self.scalar_params))
        expected_put = 0.4413197518956652
        self.assertEqual(test_put, expected_put)

    def test_price_vector_np(self):
        """Test price - np.ndarray output case"""

        # call
        test_call = self.call_opt.price(**self.vector_params)
        expected_call = np.array([[2.96746057e-01, 5.31031469e-01, 7.30298621e-01],
                                  [2.62783065e-01, 5.29285722e-01, 7.56890348e-01],
                                  [2.13141191e-01, 5.26395060e-01, 7.95937699e-01],
                                  [1.28345302e-01, 5.21278768e-01, 8.65777496e-01],
                                  [7.93566840e-04, 5.09205971e-01, 9.96790994e-01]])
        np_test.assert_allclose(test_call, expected_call)

        # put
        test_put = self.put_opt.price(**self.vector_params)
        expected_put = np.array([[0.66879322, 0.43450781, 0.23524066],
                                 [0.71099161, 0.44448895, 0.21688433],
                                 [0.7688046, 0.45555073, 0.18600809],
                                 [0.86197582, 0.46904235, 0.12454362],
                                 [0.99783751, 0.4894251, 0.00184008]])
        np_test.assert_allclose(test_put, expected_put, rtol=1e-6)

    def test_price_vector_df(self):
        """Test price - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.price(**self.vector_params)
        expected_call = pd.DataFrame(data=[[2.96746057e-01, 5.31031469e-01, 7.30298621e-01],
                                           [2.62783065e-01, 5.29285722e-01, 7.56890348e-01],
                                           [2.13141191e-01, 5.26395060e-01, 7.95937699e-01],
                                           [1.28345302e-01, 5.21278768e-01, 8.65777496e-01],
                                           [7.93566840e-04, 5.09205971e-01, 9.96790994e-01]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call)

        # put
        test_put = self.put_opt.price(**self.vector_params)
        expected_put = pd.DataFrame(data=[[0.66879322, 0.43450781, 0.23524066],
                                          [0.71099161, 0.44448895, 0.21688433],
                                          [0.7688046, 0.45555073, 0.18600809],
                                          [0.86197582, 0.46904235, 0.12454362],
                                          [0.99783751, 0.4894251, 0.00184008]],
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])

        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put)

    def test_PnL_scalar(self):
        """Test P&L - scalar case"""

        # call
        test_call = scalarize(self.call_opt.PnL(**self.scalar_params))
        expected_call = 0.23317767915072352
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.PnL(**self.scalar_params))
        expected_put = -0.22747347241997717
        self.assertEqual(test_put, expected_put)

    def test_PnL_vector_np(self):
        """Test P&L - np.ndarray output case"""

        # call
        test_call = self.call_opt.PnL(**self.vector_params)
        expected_call = np.array([[0., 0.23428541, 0.43355256],
                                  [-0.03396299, 0.23253966, 0.46014429],
                                  [-0.08360487, 0.229649, 0.49919164],
                                  [-0.16840076, 0.22453271, 0.56903144],
                                  [-0.29595249, 0.21245991, 0.70004494]])
        np_test.assert_allclose(test_call, expected_call)

        # put
        test_put = self.put_opt.PnL(**self.vector_params)
        expected_put = np.array([[0., -0.23428541, -0.43355256],
                                 [0.04219839, -0.22430427, -0.4519089],
                                 [0.10001137, -0.2132425, -0.48278514],
                                 [0.19318259, -0.19975088, -0.5442496],
                                 [0.32904428, -0.17936812, -0.66695314]])
        np_test.assert_allclose(test_put, expected_put, rtol=1e-6)

    def test_PnL_vector_df(self):
        """Test P&L - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.PnL(**self.vector_params)
        expected_call = pd.DataFrame(data=[[0., 0.23428541, 0.43355256],
                                           [-0.03396299, 0.23253966, 0.46014429],
                                           [-0.08360487, 0.229649, 0.49919164],
                                           [-0.16840076, 0.22453271, 0.56903144],
                                           [-0.29595249, 0.21245991, 0.70004494]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call)

        # put
        test_put = self.put_opt.PnL(**self.vector_params)
        expected_put = pd.DataFrame(data=[[0., -0.23428541, -0.43355256],
                                          [0.04219839, -0.22430427, -0.4519089],
                                          [0.10001137, -0.2132425, -0.48278514],
                                          [0.19318259, -0.19975088, -0.5442496],
                                          [0.32904428, -0.17936812, -0.66695314]],
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])

        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put)

    def test_delta_scalar(self):
        """Test Delta - scalar case"""

        # call
        test_call = scalarize(self.call_opt.delta(**self.scalar_params))
        expected_call = 0.025194958512498786
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.delta(**self.scalar_params))
        expected_put = copy.deepcopy(-expected_call)
        self.assertEqual(test_put, expected_put)

        # assert call and put delta consistency
        self.assertEqual(test_call, -test_put)

    def test_delta_vector_np(self):
        """Test Delta - np.ndarray output case"""

        # call
        test_call = self.call_opt.delta(**self.vector_params)
        expected_call = np.array([[0.02251146, 0.02281654, 0.01642484],
                                  [0.0245291, 0.02648423, 0.01810373],
                                  [0.02655219, 0.03231528, 0.02002786],
                                  [0.02633276, 0.0446913, 0.02109933],
                                  [0.00091364, 0.12030889, 0.00161394]])
        np_test.assert_allclose(test_call, expected_call, rtol=5e-6)

        # put
        test_put = self.put_opt.delta(**self.vector_params)
        expected_put = copy.deepcopy(-expected_call)
        np_test.assert_allclose(test_put, expected_put, rtol=5e-6)

        # assert call and put delta consistency
        np_test.assert_allclose(test_call, -test_put)

    def test_delta_vector_df(self):
        """Test Delta - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.delta(**self.vector_params)
        expected_call = pd.DataFrame(data=[[0.02251146, 0.02281654, 0.01642484],
                                           [0.0245291, 0.02648423, 0.01810373],
                                           [0.02655219, 0.03231528, 0.02002786],
                                           [0.02633276, 0.0446913, 0.02109933],
                                           [0.00091364, 0.12030889, 0.00161394]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call)

        # put
        test_put = self.put_opt.delta(**self.vector_params)
        expected_put = copy.deepcopy(-expected_call)
        pd_test.assert_frame_equal(test_put, expected_put)

        # assert call and put delta consistency
        pd_test.assert_frame_equal(test_call, -test_put)

    def test_gamma_scalar(self):
        """Test Gamma - scalar case"""

        # call
        test_call = scalarize(self.call_opt.gamma(**self.scalar_params))
        expected_call = -0.0004409117739687288
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.gamma(**self.scalar_params))
        expected_put = copy.deepcopy(-expected_call)
        self.assertEqual(test_put, expected_put)

        # assert call and put gamma coincide
        self.assertEqual(test_call, -test_put)

    def test_gamma_vector_np(self):
        """Test Gamma - np.ndarray output case"""

        # call
        test_call = self.call_opt.gamma(**self.vector_params)
        expected_call = np.array([[0.00050164, -0.00039929, -0.00076858],
                                  [0.00087371, -0.00046347, -0.00102583],
                                  [0.00161634, -0.00056552, -0.00150922],
                                  [0.0034499, -0.0007821, -0.00268525],
                                  [0.00095822, -0.00210541, -0.00130173]])
        np_test.assert_allclose(test_call, expected_call, rtol=1e-5)

        # put
        test_put = self.put_opt.gamma(**self.vector_params)
        expected_put = copy.deepcopy(-expected_call)
        np_test.assert_allclose(test_put, expected_put, rtol=1e-5)

        # assert call and put gamma coincide
        np_test.assert_allclose(test_call, -test_put)

    def test_gamma_vector_df(self):
        """Test Gamma - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.gamma(**self.vector_params)
        expected_call = pd.DataFrame(data=[[0.00050164, -0.00039929, -0.00076858],
                                           [0.00087371, -0.00046347, -0.00102583],
                                           [0.00161634, -0.00056552, -0.00150922],
                                           [0.0034499, -0.0007821, -0.00268525],
                                           [0.00095822, -0.00210541, -0.00130173]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call, check_less_precise=True)

        # put
        test_put = self.put_opt.gamma(**self.vector_params)
        expected_put = copy.deepcopy(-expected_call)
        pd_test.assert_frame_equal(test_put, expected_put, check_less_precise=True)

        # assert call and put gamma coincide
        pd_test.assert_frame_equal(test_call, -test_put)

    def test_vega_scalar(self):
        """Test Vega - scalar case"""

        # call
        test_call = scalarize(self.call_opt.vega(**self.scalar_params))
        expected_call = -0.005145983992073383
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.vega(**self.scalar_params))
        expected_put = copy.deepcopy(-expected_call)
        self.assertEqual(test_put, expected_put)

        # assert call and put vega coincide
        self.assertEqual(test_call, -test_put)

    def test_vega_vector_np(self):
        """Test Vega - np.ndarray output case"""

        # call
        test_call = self.call_opt.vega(**self.vector_params)
        expected_call = np.array([[0.00569969, -0.00560099, -0.01304515],
                                  [0.00752302, -0.00492679, -0.01319465],
                                  [0.0095413, -0.0041213, -0.01330838],
                                  [0.01087143, -0.00304268, -0.01264053],
                                  [0.00042529, -0.00115365, -0.00086306]])
        np_test.assert_allclose(test_call, expected_call, rtol=1e-5)

        # put
        test_put = self.put_opt.vega(**self.vector_params)
        expected_put = copy.deepcopy(-expected_call)
        np_test.assert_allclose(test_put, expected_put, rtol=5e-5)

        # assert call and put vega coincide
        np_test.assert_allclose(test_call, -test_put)

    def test_vega_vector_df(self):
        """Test Vega - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.vega(**self.vector_params)
        expected_call = pd.DataFrame(data=[[0.00569969, -0.00560099, -0.01304515],
                                           [0.00752302, -0.00492679, -0.01319465],
                                           [0.0095413, -0.0041213, -0.01330838],
                                           [0.01087143, -0.00304268, -0.01264053],
                                           [0.00042529, -0.00115365, -0.00086306]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call, check_less_precise=True)

        # put
        test_put = self.put_opt.vega(**self.vector_params)
        expected_put = copy.deepcopy(-expected_call)
        pd_test.assert_frame_equal(test_put, expected_put, check_less_precise=True)

        # assert call and put vega coincide
        pd_test.assert_frame_equal(test_call, -test_put)

    def test_theta_scalar(self):
        """Test Theta - scalar case"""

        # call
        test_call = scalarize(self.call_opt.theta(**self.scalar_params))
        expected_call = -3.094863279105034e-05
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.theta(**self.scalar_params))
        expected_put = 0.00016399568592748338
        self.assertEqual(test_put, expected_put)

    def test_theta_vector_np(self):
        """Test Theta - np.ndarray output case"""

        # call
        test_call = self.call_opt.theta(**self.vector_params)
        expected_call = np.array([[-4.59532646e-04, -2.10225482e-05, 3.62119705e-04],
                                  [-6.54200322e-04, -3.63343877e-05, 5.11024346e-04],
                                  [-1.01554953e-03, -6.06935961e-05, 8.07873242e-04],
                                  [-1.83825681e-03, -1.12254822e-04, 1.58102099e-03],
                                  [-4.36449645e-04, -4.24665854e-04, 9.75289889e-04]])
        np_test.assert_allclose(test_call, expected_call)

        # put
        test_put = self.put_opt.theta(**self.vector_params)
        expected_put = np.array([[0.0005918, 0.00015329, -0.00022985],
                                 [0.00078759, 0.00016973, -0.00037763],
                                 [0.00115006, 0.00019521, -0.00067336],
                                 [0.00197392, 0.00024792, -0.00144536],
                                 [0.00057325, 0.00056146, -0.00083849]])
        np_test.assert_allclose(test_put, expected_put, rtol=5e-5)

    def test_theta_vector_df(self):
        """Test Theta - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.theta(**self.vector_params)
        expected_call = pd.DataFrame(data=[[-4.59532646e-04, -2.10225482e-05, 3.62119705e-04],
                                           [-6.54200322e-04, -3.63343877e-05, 5.11024346e-04],
                                           [-1.01554953e-03, -6.06935961e-05, 8.07873242e-04],
                                           [-1.83825681e-03, -1.12254822e-04, 1.58102099e-03],
                                           [-4.36449645e-04, -4.24665854e-04, 9.75289889e-04]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call, check_less_precise=True)

        # put
        test_put = self.put_opt.theta(**self.vector_params)
        expected_put = pd.DataFrame(data=[[0.0005918, 0.00015329, -0.00022985],
                                          [0.00078759, 0.00016973, -0.00037763],
                                          [0.00115006, 0.00019521, -0.00067336],
                                          [0.00197392, 0.00024792, -0.00144536],
                                          [0.00057325, 0.00056146, -0.00083849]],
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])
        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put, check_less_precise=True)

    def test_rho_scalar(self):
        """Test Rho - scalar case"""

        # call
        test_call = scalarize(self.call_opt.rho(**self.scalar_params))
        expected_call = 0.011610379741045512
        self.assertEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.rho(**self.scalar_params))
        expected_put = -0.01727818420465756
        self.assertEqual(test_put, expected_put)

    def test_rho_vector_np(self):
        """Test Rho - np.ndarray output case"""

        # call
        test_call = self.call_opt.rho(**self.vector_params)
        expected_call = np.array([[1.21286837e-02, 1.22783358e-02, 7.54978064e-03],
                                  [1.03369366e-02, 1.12633572e-02, 6.56155667e-03],
                                  [7.93101136e-03, 9.85705868e-03, 5.12733711e-03],
                                  [4.36037779e-03, 7.67938163e-03, 2.83056260e-03],
                                  [2.23107982e-05, 3.15662549e-03, -2.24454016e-04]])
        np_test.assert_allclose(test_call, expected_call)

        # put
        test_put = self.put_opt.rho(**self.vector_params)
        expected_put = np.array([[-1.89006853e-02, -1.90503374e-02, -1.43217822e-02],
                                 [-1.55126157e-02, -1.64390362e-02, -1.17372358e-02],
                                 [-1.15090604e-02, -1.34351077e-02, -8.70538615e-03],
                                 [-6.28675585e-03, -9.60575970e-03, -4.75694066e-03],
                                 [-2.95908353e-04, -3.43022305e-03, -4.91435388e-05]])
        np_test.assert_allclose(test_put, expected_put)

    def test_rho_vector_df(self):
        """Test Theta - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.rho(**self.vector_params)
        expected_call = pd.DataFrame(data=[[1.21286837e-02, 1.22783358e-02, 7.54978064e-03],
                                           [1.03369366e-02, 1.12633572e-02, 6.56155667e-03],
                                           [7.93101136e-03, 9.85705868e-03, 5.12733711e-03],
                                           [4.36037779e-03, 7.67938163e-03, 2.83056260e-03],
                                           [2.23107982e-05, 3.15662549e-03, -2.24454016e-04]],
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call)

        # put
        test_put = self.put_opt.rho(**self.vector_params)
        expected_put = pd.DataFrame(data=[[-1.89006853e-02, -1.90503374e-02, -1.43217822e-02],
                                          [-1.55126157e-02, -1.64390362e-02, -1.17372358e-02],
                                          [-1.15090604e-02, -1.34351077e-02, -8.70538615e-03],
                                          [-6.28675585e-03, -9.60575970e-03, -4.75694066e-03],
                                          [-2.95908353e-04, -3.43022305e-03, -4.91435388e-05]],
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])
        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put)

    def test_Implied_Vol_scalar(self):
        """Test Implied Volatility - scalar case"""

        # call
        test_call = scalarize(self.call_opt.implied_volatility(**self.scalar_params))
        expected_call = 0.2
        self.assertAlmostEqual(test_call, expected_call)

        # put
        test_put = scalarize(self.put_opt.implied_volatility(**self.scalar_params))
        expected_put = 0.2
        self.assertAlmostEqual(test_put, expected_put)

    def test_Implied_Vol_vector_np(self):
        """Test Implied Volatility - np.ndarray output case"""

        # call
        test_call = self.call_opt.implied_volatility(**self.vector_params)
        expected_call = 0.2 + np.zeros_like(test_call)
        np_test.assert_allclose(test_call, expected_call, rtol=5e-7)

        # put
        test_put = self.put_opt.implied_volatility(**self.vector_params)
        expected_put = 0.2 + np.zeros_like(test_put)
        np_test.assert_allclose(test_put, expected_put, rtol=5e-7)

    def test_Implied_Vol_vector_df(self):
        """Test Implied Volatility - pd.DataFrame output case"""

        # request Pandas DataFrame as output format
        self.vector_params["np_output"] = False

        # call
        test_call = self.call_opt.implied_volatility(**self.vector_params)
        expected_call = pd.DataFrame(data=0.2 + np.zeros_like(test_call),
                                     index=self.vector_params["t"],
                                     columns=self.vector_params["S"])
        expected_call.rename_axis("S", axis='columns', inplace=True)
        expected_call.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_call, expected_call)

        # put
        test_put = self.put_opt.implied_volatility(**self.vector_params)
        expected_put = pd.DataFrame(data=0.2 + np.zeros_like(test_put),
                                    index=self.vector_params["t"],
                                    columns=self.vector_params["S"])
        expected_put.rename_axis("S", axis='columns', inplace=True)
        expected_put.rename_axis("t", axis='rows', inplace=True)
        pd_test.assert_frame_equal(test_put, expected_put)

    def test_complex_parameters_setup(self):
        """
        Test complex parameter setup:
        (S scalar, K and t vector, sigma distributed as Kxt grid, r distributed as Kxt grid)
        """

        # call
        test_call_price = self.call_opt.price(**self.complex_params)
        test_call_PnL = self.call_opt.PnL(**self.complex_params)
        test_call_delta = self.call_opt.delta(**self.complex_params)
        test_call_gamma = self.call_opt.gamma(**self.complex_params)
        test_call_vega = self.call_opt.vega(**self.complex_params)
        test_call_theta = self.call_opt.theta(**self.complex_params)
        test_call_rho = self.call_opt.rho(**self.complex_params)
        test_call_iv = self.call_opt.implied_volatility(**self.complex_params)

        expected_call_price = pd.DataFrame(data=[[0.9798568, 0.62471855, 0.47323669, 0.39201772],
                                                 [0.68363866, 0.50365944, 0.42894149, 0.37368724],
                                                 [0.87602123, 0.60825314, 0.46889718, 0.36012104]],
                                           index=self.complex_params["t"],
                                           columns=self.complex_params["K"])
        expected_call_price.rename_axis("K", axis='columns', inplace=True)
        expected_call_price.rename_axis("t", axis='rows', inplace=True)

        expected_call_PnL = pd.DataFrame(data=[[0.68311074, 0.3279725, 0.17649063, 0.09527166],
                                               [0.3868926, 0.20691339, 0.13219544, 0.07694118],
                                               [0.57927518, 0.31150708, 0.17215112, 0.06337498]],
                                         index=self.complex_params["t"],
                                         columns=self.complex_params["K"])
        expected_call_PnL.rename_axis("K", axis='columns', inplace=True)
        expected_call_PnL.rename_axis("t", axis='rows', inplace=True)

        expected_call_delta = pd.DataFrame(data=[[0.00448245, 0.02461979, 0.01726052, 0.01248533],
                                                 [0.01263986, 0.01196584, 0.01011038, 0.00854062],
                                                 [0.01504177, 0.02568663, 0.02420515, 0.02088546]],
                                           index=self.complex_params["t"],
                                           columns=self.complex_params["K"])
        expected_call_delta.rename_axis("K", axis='columns', inplace=True)
        expected_call_delta.rename_axis("t", axis='rows', inplace=True)

        expected_call_gamma = pd.DataFrame(data=[[-1.36939276e-03, -8.30886510e-04, -1.59819664e-04, -3.72062630e-05],
                                                 [-3.79395726e-04, -1.46568014e-04, -7.22169767e-05, -3.73088881e-05],
                                                 [-1.47523572e-03, -7.66683834e-04, -1.58922692e-04, 1.82659675e-04]],
                                           index=self.complex_params["t"],
                                           columns=self.complex_params["K"])
        expected_call_gamma.rename_axis("K", axis='columns', inplace=True)
        expected_call_gamma.rename_axis("t", axis='rows', inplace=True)

        expected_call_vega = pd.DataFrame(data=[[-0.00777965, -0.00944069, -0.00272385, -0.00084549],
                                                [-0.00559895, -0.00259558, -0.00149204, -0.00088094],
                                                [-0.00294643, -0.00170141, -0.00038795, 0.00048643]],
                                          index=self.complex_params["t"],
                                          columns=self.complex_params["K"])
        expected_call_vega.rename_axis("K", axis='columns', inplace=True)
        expected_call_vega.rename_axis("t", axis='rows', inplace=True)

        expected_call_theta = pd.DataFrame(data=[[1.67739087e-04, 2.81595503e-04, 7.08163129e-05, -1.41282958e-05],
                                                 [9.90248652e-04, 4.91233384e-04, 3.00397593e-04, 1.78375694e-04],
                                                 [1.31411352e-02, 8.04031544e-03, 1.61848869e-03, -3.41813594e-03]],
                                           index=self.complex_params["t"],
                                           columns=self.complex_params["K"])
        expected_call_theta.rename_axis("K", axis='columns', inplace=True)
        expected_call_theta.rename_axis("t", axis='rows', inplace=True)

        expected_call_rho = pd.DataFrame(data=[[-0.00404295, 0.01115923, 0.00757627, 0.00513166],
                                               [0.00165411, 0.00208889, 0.00175266, 0.0014392],
                                               [0.00013089, 0.00046672, 0.00046837, 0.00041632]],
                                         index=self.complex_params["t"],
                                         columns=self.complex_params["K"])
        expected_call_rho.rename_axis("K", axis='columns', inplace=True)
        expected_call_rho.rename_axis("t", axis='rows', inplace=True)

        expected_call_iv = pd.DataFrame(data=self.complex_params["sigma"],
                                        index=self.complex_params["t"],
                                        columns=self.complex_params["K"])
        expected_call_iv.rename_axis("K", axis='columns', inplace=True)
        expected_call_iv.rename_axis("t", axis='rows', inplace=True)

        pd_test.assert_frame_equal(test_call_price, expected_call_price)
        pd_test.assert_frame_equal(test_call_PnL, expected_call_PnL)
        pd_test.assert_frame_equal(test_call_delta, expected_call_delta)
        pd_test.assert_frame_equal(test_call_gamma, expected_call_gamma)
        pd_test.assert_frame_equal(test_call_vega, expected_call_vega, check_less_precise=True)
        pd_test.assert_frame_equal(test_call_theta, expected_call_theta)
        pd_test.assert_frame_equal(test_call_rho, expected_call_rho, check_less_precise=True)
        pd_test.assert_frame_equal(test_call_iv.iloc[:, :-1], expected_call_iv.iloc[:, :-1])
        self.assertAlmostEqual(test_call_iv.iloc[-1, -1], expected_call_iv.iloc[-1, -1], places=5)

        # put
        test_put_price = self.put_opt.price(**self.complex_params)
        test_put_PnL = self.put_opt.PnL(**self.complex_params)
        test_put_delta = self.put_opt.delta(**self.complex_params)
        test_put_gamma = self.put_opt.gamma(**self.complex_params)
        test_put_vega = self.put_opt.vega(**self.complex_params)
        test_put_theta = self.put_opt.theta(**self.complex_params)
        test_put_rho = self.put_opt.rho(**self.complex_params)
        test_put_iv = self.put_opt.implied_volatility(**self.complex_params)

        expected_put_price = pd.DataFrame(data=[[0.01315404, 0.36135198, 0.50594203, 0.58031737],
                                                [0.29830713, 0.47471481, 0.54587421, 0.59758286],
                                                [0.12151605, 0.38901089, 0.52809366, 0.63659669]],
                                          index=self.complex_params["t"],
                                          columns=self.complex_params["K"])
        expected_put_price.rename_axis("K", axis='columns', inplace=True)
        expected_put_price.rename_axis("t", axis='rows', inplace=True)

        expected_put_PnL = pd.DataFrame(data=[[-0.65563919, -0.30744125, -0.16285119, -0.08847586],
                                              [-0.37048609, -0.19407842, -0.12291901, -0.07121037],
                                              [-0.54727717, -0.27978234, -0.14069957, -0.03219654]],
                                        index=self.complex_params["t"],
                                        columns=self.complex_params["K"])
        expected_put_PnL.rename_axis("K", axis='columns', inplace=True)
        expected_put_PnL.rename_axis("t", axis='rows', inplace=True)

        expected_put_delta = copy.deepcopy(-expected_call_delta)

        expected_put_gamma = copy.deepcopy(-expected_call_gamma)

        expected_put_vega = copy.deepcopy(-expected_call_vega)

        expected_put_theta = pd.DataFrame(data=[[-1.40533310e-04, -2.27564242e-04, 9.66413009e-06, 1.20685566e-04],
                                                [-8.55735531e-04, -3.30404740e-04, -1.13446636e-04, 3.45054237e-05],
                                                [-1.28951671e-02, -7.76709242e-03, -1.31802569e-03, 3.74582396e-03]],
                                          index=self.complex_params["t"],
                                          columns=self.complex_params["K"])
        expected_put_theta.rename_axis("K", axis='columns', inplace=True)
        expected_put_theta.rename_axis("t", axis='rows', inplace=True)

        expected_put_rho = pd.DataFrame(data=[[-0.00292173, -0.01807524, -0.01444393, -0.01195132],
                                              [-0.00523216, -0.00565392, -0.00530473, -0.00497835],
                                              [-0.00040418, -0.00073995, -0.00074152, -0.00068939]],
                                        index=self.complex_params["t"],
                                        columns=self.complex_params["K"])
        expected_put_rho.rename_axis("K", axis='columns', inplace=True)
        expected_put_rho.rename_axis("t", axis='rows', inplace=True)

        expected_put_iv = pd.DataFrame(data=self.complex_params["sigma"],
                                       index=self.complex_params["t"],
                                       columns=self.complex_params["K"])
        expected_put_iv.rename_axis("K", axis='columns', inplace=True)
        expected_put_iv.rename_axis("t", axis='rows', inplace=True)

        pd_test.assert_frame_equal(test_put_price, expected_put_price)
        pd_test.assert_frame_equal(test_put_PnL, expected_put_PnL)
        pd_test.assert_frame_equal(test_put_delta, expected_put_delta)
        pd_test.assert_frame_equal(test_put_gamma, expected_put_gamma)
        pd_test.assert_frame_equal(test_put_vega, expected_put_vega, check_less_precise=True)
        pd_test.assert_frame_equal(test_put_theta, expected_put_theta)
        pd_test.assert_frame_equal(test_put_rho, expected_put_rho, check_less_precise=True)
        pd_test.assert_frame_equal(test_put_iv.iloc[:, :-1], expected_put_iv.iloc[:, :-1])
        self.assertAlmostEqual(test_put_iv.iloc[-1, -1], expected_put_iv.iloc[-1, -1], places=5)

        # test gamma and vega consistency
        pd_test.assert_frame_equal(test_call_delta, -test_put_delta)
        pd_test.assert_frame_equal(test_call_gamma, -test_put_gamma)
        pd_test.assert_frame_equal(test_call_vega, -test_put_vega)


if __name__ == '__main__':
    unittest.main()



"""
Created by: Gabriele Pompa (gabriele.pompa@gmail.com)

File: portfolio.py

Created on Tue Jul 14 2020 - Version: 1.0

Description: 
    
This file contains the definition of Portfolio class.
"""

# ----------------------- standard imports ---------------------------------- #
# for NumPy arrays
import numpy as np

# ----------------------- sub-modules imports ------------------------------- #

from ..utils.utils import *


# -----------------------------------------------------------------------------#

class Portfolio:
    """
    Portfolio class modeling a portfolio of options. It implements an aggregation of EuropeanOption sub-classes 
    defining the .add_instrument() composition method which takes in input a FinancialInstrument (either
    PlainVanillaOption or DigitalOption) to be added to the portfolio.
    
    Attributes:
    -----------
    
        composition (List of Dicts): List of Dicts, each describing a single constituent FinancialInstrument, together
                                     with the position the portfolio is holding on it.
        info and mkt_info (Strings): information labels on portfolio and constituent instruments.
        S (Float):                   underlying value when the portfolio is formed.
        K (np.ndarray):              Strikes of constituent options.
        tau (np.ndarray):            Time(s) to maturity of constituent options, when the portfolio is formed.
        is_multi_horizon (Bool):     True if constituent options have different expiration dates.
        
    Public Methods:
    --------
    
        getters for all attributes
        
        setters for common attributes, not belonging to mkt_env
        
        payoff: float
            Computes the payoff of the portfolio.

        price: float
            Computes the Black-Scholes value of the portfolio.

        PnL: float
            Computes the P&L of the portfolio.
            
        delta: float
            Computes the Black-Scholes delta of the portfolio.

        theta: float
            Computes the Black-Scholes theta of the portfolio.

        gamma: float
            Computes the Black-Scholes gamma of the portfolio.

        vega: float
            Computes the Black-Scholes vega of the portfolio.

        rho: float
            Computes the Black-Scholes rho of the portfolio.

    Instantiation and Usage examples: 
    --------   
        
        - portfolio.py
        - portfolio_single_strike.py
        - example_portfolio_multi_strike.py
        - example_portfolio_bull_spread.py
        - example_portfolio_bull_spread_other_params.py
        - example_portfolio_calendar_spread.py
        - example_portfolio_calendar_spread_other_params.py

    """

    def __init__(self, name="Dummy"):

        # initialize an empty portfolio
        self.__composition = []

        # initialize empty info strings
        self.__info = "{} Portfolio: \n".format(name)
        self.__mkt_info = None

        # initialize valuation date, underlying value, strikes and times-to-maturity attributes
        self.__t = None
        self.__T = np.array([])
        self.__S = None
        self.__sigma = None
        self.__r = None
        self.__K = np.array([])
        self.__tau = np.array([])
        self.is_multi_horizon = False
        self.is_multi_strike = False
        self.is_empty = True

    def __repr__(self):
        return self.get_info()

    # 
    # getters
    #

    def get_info(self):
        return self.__info

    def get_mkt_info(self):
        return self.__mkt_info

    def get_t(self):
        return self.__t

    def get_T(self):
        return scalarize(self.__T)

    def get_K(self):
        return self.__K

    def get_S(self):
        return self.__S

    def get_sigma(self):
        return self.__sigma

    def get_r(self):
        return self.__r

    def get_tau(self):
        return self.__tau

    def get_composition(self):
        return self.__composition

    #
    # setters
    #

    def set_t(self, t):
        self.__t = t

    def set_S(self, S):
        self.__S = S

    def set_sigma(self, sigma):
        self.__sigma = sigma

    def set_r(self, r):
        self.__r = r

    #
    # Composition method
    #

    def add_instrument(self, FinancialInstrument, position):

        if self.is_empty:
            self.is_empty = False

        long_short = 'Long' if position > 0 else 'Short'
        instrument_info = long_short + " {} ".format(abs(position)) + FinancialInstrument.get_info()

        self.__composition.append({"instrument": FinancialInstrument,
                                   "position": position,
                                   "info": instrument_info})

        # update portfolio info strings
        self.__update_info(FinancialInstrument)

        # update portfolio attributes
        self.__update_t(FinancialInstrument)
        self.__update_T(FinancialInstrument)
        self.__update_S(FinancialInstrument)
        self.__update_sigma(FinancialInstrument)
        self.__update_r(FinancialInstrument)
        self.__update_K(FinancialInstrument)
        self.__update_tau(FinancialInstrument)

    # 
    # Private methods
    #

    def __update_info(self, fin_inst):
        self.__info += self.__composition[-1]["info"] + "\n"
        if self.__mkt_info is None:
            self.__mkt_info = fin_inst.get_mkt_info()

    def __update_t(self, fin_inst):
        if self.get_t() is None:
            self.set_t(fin_inst.get_t())
        else:
            if self.get_t() != fin_inst.get_t():
                raise ValueError(
                    "No multiple valuation dates in input allowed: \n\n current: {}, \n\n other given input: {}"
                    .format(self, self.get_t(), fin_inst.get_t()))

    def __update_T(self, fin_inst):
        expiration_dates = np.append(self.get_T(), fin_inst.get_T())
        # filter only distinct strikes
        self.__T = iterable_to_numpy_array(np.unique(expiration_dates), sort_func=date_string_to_datetime_obj)
        # check if the portfolio is a multi-horizon portfolio
        if len(self.__T) > 1:
            self.is_multi_horizon = True

    def __update_S(self, fin_inst):
        if self.get_S() is None:
            self.set_S(fin_inst.get_S())

    def __update_sigma(self, fin_inst):
        if self.get_sigma() is None:
            self.set_sigma(fin_inst.get_sigma())

    def __update_r(self, fin_inst):
        if self.get_r() is None:
            self.set_r(fin_inst.get_r())

    def __update_K(self, fin_inst):
        # append new instrument strike
        strikes = np.append(self.get_K(), fin_inst.get_K())
        # filter only distinct strikes
        self.__K = np.unique(strikes)
        # check if the portfolio is a multi-strike portfolio
        if len(self.__K) > 1:
            self.is_multi_strike = True
        else:
            self.__K = self.__K[0]

    def __update_tau(self, fin_inst):
        # append new instrument tau
        times_to_maturity = np.append(self.get_tau(), fin_inst.get_tau())
        # filter only distinct times-to-maturity
        self.__tau = np.unique(times_to_maturity)
        # consistency check
        if (len(self.__tau) > 1) and (not self.is_multi_horizon):
            raise AttributeError("Multi-horizon portfolio not properly handled: \n \tau = {}"
                                 .format(self.__tau))

    def check_parameters(self, *args, **kwargs):
        """Check both x-axis and time dimensional parameters."""

        # check x-axis
        self.__check_x_axis(**kwargs)

        # check time parameter
        self.__check_time_parameter(*args, **kwargs)

    def __check_x_axis(self, **kwargs):
        """Check that multi-strike portfolio do not get strike as input x-axis parameter,
        which is not well defined."""

        # x-axis parameter:
        strike = kwargs['K'] if 'K' in kwargs else None

        if self.is_multi_strike and strike is not None:
            raise NotImplementedError("No 'strike' x-axis parameter allowed for multi-strike portfolio.")

    def __check_time_parameter(self, *args, **kwargs):
        """Check that multi-horizon portfolio do not get time(s)-to-maturity as input time parameter,
        which is not well defined."""

        # time parameter:
        time_param = args[1] if len(args) > 1 \
            else kwargs['tau'] if 'tau' in kwargs \
            else (kwargs['t'] if 't' in kwargs else None)

        # Case of no time parameter in input allowed: sigma x r grid case
        if time_param is not None:

            # check that time parameter is not a time-to-maturity if the portfolio is multi-horizon:
            if self.is_multi_horizon and is_numeric(time_param):
                raise TypeError(
                    "No time-to-maturity time parameter allowed for multi-horizon portfolio: \n\n tau={} given in input"
                    .format(time_param))

                #

    # Public methods
    #

    def time_to_maturity(self, *args, **kwargs):
        """
        Utility method to compute time-to-maturity of the portfolio, 
        if possible (that is, if the portfolio is not multi-horizon and if it's
        not empty. Method taken from an instrument.
        """

        if self.is_empty:
            raise NotImplementedError("No time-to-maturity defined for empty portfolio")
        elif self.is_multi_horizon:
            raise NotImplementedError("No time-to-maturity defined for multi-horizon portfolio")
        else:
            return self.get_composition()[0]["instrument"].time_to_maturity(*args, **kwargs)

    def payoff(self, *args, **kwargs):
        """
        Returns the portfolio payoff as the scalar product (i.e. sum of elementwise products) 
        between single instrument payoffs and positions.
        
        Can be called with the same signature of the .payoff() public method of
        constituent options.
        """

        # check parameters
        self.check_parameters(*args, **kwargs)

        # portfolio payoff is the sum position * instrument_payoff
        return sum([inst["position"] * inst["instrument"].payoff(*args, **kwargs) for inst in self.get_composition()])

    def price(self, *args, **kwargs):
        """
        Returns the portfolio value as the scalar product (i.e. sum of elementwise products) 
        between single instrument prices and positions.
        
        Can be called with the same signature of the .price() public method of
        constituent options.
        """

        # check parameters
        self.check_parameters(*args, **kwargs)

        # portfolio price is the sum position * instrument_price
        return sum([inst["position"] * inst["instrument"].price(*args, **kwargs) for inst in self.get_composition()])

    def PnL(self, *args, **kwargs):
        """
        Returns the portfolio Profit & Loss as the scalar product (i.e. sum of elementwise products) 
        between single instrument P&Ls and positions.
        
        Can be called with the same signature of the .PnL() public method of
        constituent options.
        """

        # check parameters
        self.check_parameters(*args, **kwargs)

        # portfolio P&L is the sum position * instrument_payoff
        return sum([inst["position"] * inst["instrument"].PnL(*args, **kwargs) for inst in self.get_composition()])

    def delta(self, *args, **kwargs):
        """
        Returns the portfolio Delta as the scalar product (i.e. sum of elementwise products) 
        between single instrument Deltas and positions.
        
        Can be called with the same signature of the .delta() public method of
        constituent options.
        """

        # check parameters
        self.check_parameters(*args, **kwargs)

        # portfolio delta is the sum position * instrument_payoff
        return sum([inst["position"] * inst["instrument"].delta(*args, **kwargs) for inst in self.get_composition()])

    def theta(self, *args, **kwargs):
        """
        Returns the portfolio Theta as the scalar product (i.e. sum of elementwise products) 
        between single instrument Thetas and positions.
        
        Can be called with the same signature of the .theta() public method of
        constituent options.
        """

        # check parameters
        self.check_parameters(*args, **kwargs)

        # portfolio theta is the sum position * instrument_payoff
        return sum([inst["position"] * inst["instrument"].theta(*args, **kwargs) for inst in self.get_composition()])

    def gamma(self, *args, **kwargs):
        """
        Returns the portfolio Gamma as the scalar product (i.e. sum of elementwise products) 
        between single instrument Gammas and positions.
        
        Can be called with the same signature of the .gamma() public method of
        constituent options.
        """

        # check parameters
        self.check_parameters(*args, **kwargs)

        # portfolio gamma is the sum position * instrument_payoff
        return sum([inst["position"] * inst["instrument"].gamma(*args, **kwargs) for inst in self.get_composition()])

    def vega(self, *args, **kwargs):
        """
        Returns the portfolio Vega as the scalar product (i.e. sum of elementwise products) 
        between single instrument Vegas and positions.
        
        Can be called with the same signature of the .vega() public method of
        constituent options.
        """

        # check parameters
        self.check_parameters(*args, **kwargs)

        # portfolio vega is the sum position * instrument_payoff
        return sum([inst["position"] * inst["instrument"].vega(*args, **kwargs) for inst in self.get_composition()])

    def rho(self, *args, **kwargs):
        """
        Returns the portfolio Rho as the scalar product (i.e. sum of elementwise products) 
        between single instrument Rhos and positions.
        
        Can be called with the same signature of the .rho() public method of
        constituent options.
        """

        # check parameters
        self.check_parameters(*args, **kwargs)

        # portfolio rho is the sum position * instrument_payoff
        return sum([inst["position"] * inst["instrument"].rho(*args, **kwargs) for inst in self.get_composition()])



"""
Created by: Gabriele Pompa (gabriele.pompa@gmail.com)

File: portfolio.py

Created on Tue Jul 14 2020 - Version: 1.0

Description: 
    
This file contains definitions for general utility functions.
"""

# ----------------------- standard imports ---------------------------------- #
# for NumPy arrays
import numpy as np

# for Pandas Series and DataFrame
import pandas as pd

# for date management
import datetime as dt

# for Matplotlib plotting
import matplotlib.pyplot as plt

# to handle dates in Matplotlib
import matplotlib.dates as mpl_dates

# to identify iterable data-structures
from collections.abc import Iterable


# -----------------------------------------------------------------------------#

def scalarize(x):
    """
    Reduce array x to scalar, if possible.
    """

    if is_iterable_not_string(x) and x.size == 1:
        return x.item()
    else:
        return x


# -----------------------------------------------------------------------------#

def iterable_to_numpy_array(x, sort=True, sort_func=None, reverse_order=False):
    """
    Create a 1-dim NumPy Array from a 1-dim Iterable with elements of the same type.
    If the iterable has elements of different data-type, it raises an error.
    
    If sort is True (default), returns a sorted array. Optionally uses a custom 
    function sort_func to sort the iterable (default: ascending order).
    """

    if is_iterable_not_string(x) and test_same_type(x):
        if not isinstance(x, np.ndarray):
            x = np.array([xi for xi in x])
        return np.array(sorted(x, key=sort_func, reverse=reverse_order)) if sort else x
    else:
        return x


# -----------------------------------------------------------------------------#

def homogenize(x, *args, **kwargs):
    """
    Utility function to homogenize variable x, calling:
        - function iterable_to_numpy_array()
        - function scalarize()
    sequentially.
    """

    # convert x into a NumPy array if it is an Iterable of same data-type
    x = iterable_to_numpy_array(x, *args, **kwargs)

    # reduce x to a scalar, if possible
    x = scalarize(x)

    return x


# -----------------------------------------------------------------------------#

def coordinate(x, y, *args, x_name="x", y_name="y", others_scalar={}, others_vector={}, np_output=True, **kwargs):
    """
    Utility function to coordinate two main parameters x and y, each other and
    possibly with others scalar parameters scalar or vector.
    
    Any other vector parameter p should be of the same shape of coord_x, that is
    
        - if x is m-shaped and y is n-shaped: p should be (m,n) shaped (or reshapable);
        - if x is m-shaped and y is scalar: p should be (m,) shaped (or reshapable);
        - if x is scalar and y is n-shaped: p should be (n,) shaped (or reshapable);
        - if x and y are scalar: p should be (1,) shaped (or reshapable).
    """

    # test dimensionality: x and y can either be scalar or 1-dim arrays
    if is_iterable(x):
        test_dim(x, dim=1)
    if is_iterable(y):
        test_dim(y, dim=1)

    # coordinate the two main parameters (x,y) ---> (coord_x, coord_y)
    coord_x, coord_y = coordinate_x_and_y(x, y, *args, np_output=np_output,
                                          x_name=x_name, y_name=y_name, **kwargs)

    # initialize output dictionary of coordinated parameters
    coordinated_parameters = {x_name: coord_x, y_name: coord_y}

    # coordinate each other scalar parameter p (if any) with coord_x: p ---> coord_p
    for p_name in others_scalar:

        p = others_scalar[p_name]

        coord_p_scal = coordinate_y_with_x(x=coord_x, y=p, np_output=np_output)

        if p_name not in coordinated_parameters:
            coordinated_parameters[p_name] = coord_p_scal
        else:
            raise KeyError("Duplicate parameter name: {}".format(p_name))

    # coordinate each other vector parameter p_vec (if any) with coord_x
    for p_name in others_vector:

        coord_p_vec = others_vector[p_name]

        # if different shapes, try reshaping
        if coord_p_vec.shape != coord_x.shape:
            try:
                coord_p_vec = coord_p_vec.reshape(coord_x.shape)
            except ValueError:
                raise ValueError("Cannot reshape parameter '{}' of shape {} into shape {} of parameter '{}'"
                                 .format(p_name, coord_p_vec.shape, coord_x.shape, x_name))

        # if parameters should be pd.DataFrame, recast
        if (not np_output) and (not isinstance(coord_p_vec, pd.DataFrame)):
            coord_p_vec = coordinate_y_with_x(x=coord_x,
                                              y=coord_p_vec,
                                              np_output=np_output)

        if p_name not in coordinated_parameters:
            coordinated_parameters[p_name] = coord_p_vec
        else:
            raise KeyError("Duplicate parameter name: {}".format(p_name))

    # return output dictionary of coordinated parameters
    return coordinated_parameters


# -----------------------------------------------------------------------------#

def coordinate_y_with_x(x, y, np_output):
    """
    Utility function to coordinate the scalar/vector parameter y with a 
    np.ndarray/pd.DataFrame x. If y is a vector, it must be of the same shape 
    of x. We distinguish the two cases according to the value of the boolean 
    flag np_output:
        
        - If np_output is True, x is expected to be a np.ndarray and y will be returned 
          as a np.ndarray x-shaped filled with y value(s).
          
        - If np_output is False, x is expected to be a pd.DataFrame and y will be returned 
          as a pd.DataFrame identical to x filled with y value(s).
    """
    if np_output:
        if isinstance(x, np.ndarray):
            y_coord_x = y + np.zeros_like(x)
        else:
            raise TypeError(
                r"Inconsistent type of \n x={} \n parameter in input: \n type(x)={} (np.ndarray expected)"
                .format(x, type(x)))
    else:
        if isinstance(x, pd.DataFrame):
            y_coord_x = pd.DataFrame(data=y, index=x.index, columns=x.columns)

        else:
            raise TypeError(
                r"Inconsistent type of \n x={} \n parameter in input: \n type(x)={} (pd.DataFrame expected)"
                .format(x, type(x)))
    return y_coord_x


# -----------------------------------------------------------------------------#

def coordinate_x_and_y(x, y, *args, np_output, **kwargs):
    """
    Utility function to coordinate the two scalar/np.ndarray variables x and y:
        
        - as NumPy Arrays: calling coordinate_as_ndarray() function
        - as Pandas DataFrames: calling coordinate_as_df() function
    """

    if np_output:
        return coordinate_x_and_y_as_ndarray(x, y)
    else:
        return coordinate_x_and_y_as_df(x, y, *args, **kwargs)


# -----------------------------------------------------------------------------#

def coordinate_x_and_y_as_ndarray(x, y):
    """
    Utility function to coordinate the two scalar/np.ndarray variables x and y
    as NumPy Arrays. The following cases are considered:
        
        1) if x is array of length n; y is array of length m, then:
            x, y ---> (m, n) shaped arrays creating a mesh-grid
            (see np.meshgrid documentation)
            
        2) if x is array of length n; y is scalar, then:
            y ---> array of length n, repeating its value n-times

        3) if y is array of length m; x is scalar, then:
            x ---> array of length m, repeating its value m-times
        
        4) if both x and y are scalar, then:
            y, x ---> array of length 1 made of their own values
    """

    if is_iterable(x) and is_iterable(y):
        # case 1

        # creating a mesh-grid combining x and y
        x, y = np.meshgrid(x, y)

    elif is_iterable(x) and (not is_iterable(y)):
        # case 2

        # length of x
        n = len(x)

        # make y look like x
        y = np.repeat(y, repeats=n)

    elif (not is_iterable(x)) and is_iterable(y):
        # case 3

        # length of y
        m = len(y)

        # make x look like y
        x = np.repeat(x, repeats=m)

    else:
        # case 4 

        # make x and y length-1 arrays
        x = np.array([x])
        y = np.array([y])

    return x, y


# -----------------------------------------------------------------------------#

def coordinate_x_and_y_as_df(x, y, col_labels, ind_labels, **kwargs):
    """
    Utility function to coordinate the two scalar/np.ndarray variables x and y
    as Pandas DataFrames. The following cases are considered:
        
        1) if x has length n and y has length m, then:
            x, y ---> [m rows x n cols] pd.DataFrame defining a mesh-grid
            of coordinates (see np.meshgrid documentation)
            
        2) if x has length n and y is a scalar, then:
            x ---> [1 row x n cols] pd.DataFrame with x as values
            y ---> [1 row x n cols] pd.DataFrame with y value repeated n-times along the row

        3) if x is a scalar and y has length m, then:
            x ---> [m rows x 1 col] pd.DataFrames with x value repeated m-times along the column
            y ---> [m rows x 1 col] pd.DataFrame with y as values
        
        4) if both x and y are scalar, then:
            y, x ---> [1 row x 1 col] pd.DataFrame with x and y value, respectively
    
    Scalar/Iterables col_labels and ind_labels are used to set the columns and indexes
    of the returned dataframes.
    
    Parameters:
        
        x (scalar; np.ndarray): x variable
        y (scalar; np.ndarray): y variable
        col_labels (scalar or Iterable): defines the column labels
        ind_labels (scalar or Iterable): defines the indexes
        

    Results:
        
        x_df (pd.DataFrame): homogenized x variable
        y_df (pd.DataFrame): homogenized y variable
    """

    if is_iterable(x) and is_iterable(y):
        # case 1    

        # creating a mesh-grid combining x and y
        x, y = np.meshgrid(x, y)

    elif is_iterable(x) and (not is_iterable(y)):
        # case 2

        # length of x
        n = len(x)

        # make y look like x
        y = np.repeat(y, repeats=n)

        # reshape x and y to shape (1,n)
        x = x.reshape((1, n))
        y = y.reshape((1, n))

    elif (not is_iterable(x)) and is_iterable(y):
        # case 3

        # length of y
        m = len(y)

        # make x look like y
        x = np.repeat(x, repeats=m)

    else:
        # case 4 

        # make x and y length-1 arrays
        x = np.array([x])
        y = np.array([y])

    # set labels for columns and indexes
    cols = col_labels if is_iterable_not_string(col_labels) else np.array([col_labels])
    inds = ind_labels if is_iterable_not_string(ind_labels) else np.array([ind_labels])

    x_df = pd.DataFrame(data=x,
                        index=inds,
                        columns=cols)

    y_df = pd.DataFrame(data=y,
                        index=inds,
                        columns=cols)

    # optional renaming of the column axis
    if 'x_name' in kwargs:
        x_df.rename_axis(kwargs['x_name'], axis='columns', inplace=True)
        y_df.rename_axis(kwargs['x_name'], axis='columns', inplace=True)

        # optional renaming of the index axis
    if 'y_name' in kwargs:
        x_df.rename_axis(kwargs['y_name'], axis='rows', inplace=True)
        y_df.rename_axis(kwargs['y_name'], axis='rows', inplace=True)

    return x_df, y_df


# -----------------------------------------------------------------------------#

def test_dim(iterable_obj, dim=1):
    """
    Utility function to test whether an iterable_obj is of dimension dim,
    checking its .ndim attribute.
    """

    try:
        if iterable_obj.ndim != dim:
            raise ValueError("Iterable obj: {} has dimension={}; expected dimension={}"
                             .format(iterable_obj, iterable_obj.ndim, dim))
    except AttributeError:
        raise AttributeError("Iterable obj: {} of type {} has no attribute 'ndim'"
                             .format(iterable_obj, type(iterable_obj)))


# -----------------------------------------------------------------------------#

def test_same_type(iterable_obj):
    """
    Utility function to test whether all elements of an iterable_obj are of the 
    same type. If not it raises a TypeError.
    """
    # by set definition, the set of types of the elements in iterable_obj
    # includes all and only the different types of the elements in iterable_obj.
    # If its length is 1, then all the elements of iterable_obj are of the 
    # same data-type
    if len(set([type(x) for x in iterable_obj])) == 1:
        # all element are of the same type: test successful!
        return True
    else:
        raise TypeError("Iterable '{}' in input has elements of heterogeneous types: {}"
                        .format(iterable_obj, [type(x) for x in iterable_obj]))


# -----------------------------------------------------------------------------#

def test_valid_format(date_string, date_format="%d-%m-%Y"):
    """
    Utility function to test whether:
        
        1-dim case:
            a date_string String is  
            
        Multi-dim case:
            a (non-String) Iterable has elements
        
    conform to the date_format (default: 'dd-mm-YYYY') date format. 
    If not, it raises a ValueError.
    
    If date_string in input is neither an Iterable, nor a String, it raises a TypeError.
    """

    try:
        if isinstance(date_string, str):
            # 1-dim case
            dt.datetime.strptime(date_string, date_format)
        elif is_iterable_not_string(date_string):
            # Multi-dim case
            pd.to_datetime(date_string, format=date_format, errors='raise')
        else:
            # neither an Iterable, nor a String: raise TypeError
            raise TypeError("Type {} of date_string {} not recognized".format(type(date_string), date_string))

    except ValueError:
        # not conform to date_format: raise ValueError
        raise ValueError("date_string {} in input is not conform to 'dd-mm-YYYY' date format".format(date_string))
    else:
        # conform to date_format: test successful!
        return True


# -----------------------------------------------------------------------------#

def datetime_obj_to_date_string(date):
    """
    Utility function to convert: 
        
        1-dim case:
            from dt.datetime object --> to 'dd-mm-YYYY' String
        
        Multi-dim case:
        
            from pd.DatetimeIndex --> to pd.Index of 'dd-mm-YYYY' String
            from Iterable --> to List of 'dd-mm-YYYY' String
    """

    if isinstance(date, dt.datetime) or isinstance(date, pd.DatetimeIndex):
        # .strftime() is a polymorphic method, implemented by both 
        # datetime objects of datetime (1-dim) and DatetimeIndex (Multi-dim) objects of Pandas 
        # so there is no need to differentiate between the two case when calling it
        return date.strftime("%d-%m-%Y")
    elif is_iterable_not_string(date):
        # all other kind of iterables (Lists, np.ndarray, etc..) are mapped to Lists
        return [d.strftime("%d-%m-%Y") for d in date]
    else:
        return date


# -----------------------------------------------------------------------------#

def date_string_to_datetime_obj(date_string):
    """
    Utility function to convert: 
        
        1-dim case:
            from String object conform to 'dd-mm-YYYY' date format --> to dt.datetime.
        
        Multi-dim case:
            from (non-String) Iterable objects of elements conform to 'dd-mm-YYYY' date format --> to pd.DatetimeIndex
            
    The 'dd-mm-YYYY' date format is controlled thought test_valid_format() utility function.
    """

    if isinstance(date_string, str) and test_valid_format(date_string):
        # 1-dim case
        return dt.datetime.strptime(date_string, "%d-%m-%Y")
    elif is_iterable_not_string(date_string) and test_valid_format(date_string):
        # Multi-dim case
        return pd.DatetimeIndex(date_string)
    else:
        return date_string


# -----------------------------------------------------------------------------#

def date_to_number(date):
    """
    Utility function to convert a date-like object into its numeric representation.
    Useful in matplotlib plots with dates axes. AttributeError handled.
    """

    try:
        return mpl_dates.date2num(date) if is_date(date) else date
    except AttributeError:
        raise


# -----------------------------------------------------------------------------#

def is_iterable(x):
    """
    Utility function to check if input can be iterated over (that is, if input is a List, np.array, pd.date_range, etc.)
    """
    return isinstance(x, Iterable)


# -----------------------------------------------------------------------------#

def is_iterable_not_string(x):
    """
    Utility function to check if input can be iterated over (that is, if input is a List, np.array, pd.date_range, etc.)
    but it is not a String
    """
    return is_iterable(x) and not isinstance(x, str)


# -----------------------------------------------------------------------------#

def is_numeric(x):
    """
    Utility function to check if input is/contains numeric data.
    """

    if is_iterable_not_string(x) and test_same_type(x):
        # since all elements are of the same type, 
        # it's enough to check the first element
        return isinstance(x[0], float) or isinstance(x[0], int)
    else:
        return isinstance(x, float) or isinstance(x, int)


# -----------------------------------------------------------------------------#

def is_date(x):
    """
    Utility function to check if input is/contains date-like data.
    The error due to invalid (non 'dd-mm-YYYY') date Strings is controlled thanks to test_valid_format() function.
    """

    if is_iterable_not_string(x) and test_same_type(x):
        # since all elements are of the same type, 
        # it's enough to check the first element
        return isinstance(x[0], dt.datetime) or (isinstance(x[0], str) and test_valid_format(x[0]))
    else:
        return isinstance(x, dt.datetime) or (isinstance(x, str) and test_valid_format(x))


# -----------------------------------------------------------------------------#

def plot_compare(x, f, f_ref, **kwargs):
    """
    Plotting function to compare a function f(x) with another reference function
    f_ref(x). It makes 6 plots:
        
        [Top-Left]     f(x) Vs x    
        [Top-Right]    f_ref(x) Vs x
        [Mid-Left]     f(x) - f_ref(x) Vs x
        [Mid-Right]    (f(x) - f_ref(x)) / f_ref(x) Vs x
        [Bottom-Left]  |f(x) - f_ref(x)| Vs x
        [Bottom-Right] |(f(x) - f_ref(x)) / f_ref(x)| Vs x
    """

    # parsing optional parameters
    f_label = kwargs['f_label'] if 'f_label' in kwargs else "f"
    f_ref_label = kwargs['f_ref_label'] if 'f_ref_label' in kwargs else "f_ref"
    title = kwargs['title'] if 'title' in kwargs else "f Vs f_ref comparison"
    x_label = kwargs['x_label'] if 'x_label' in kwargs else "x"
    top_left_subtitle = kwargs['f_test_name'] if 'f_test_name' in kwargs else "Test function"
    top_right_subtitle = kwargs['f_ref_name'] if 'f_ref_name' in kwargs else "Reference function"

    # define the figure
    fig, axs = plt.subplots(figsize=(17, 10), nrows=3, ncols=2)

    # [Top-Left] f(x) Vs x
    axs[0, 0].plot(x, f, 'b-', lw=1.5)
    axs[0, 0].set_ylabel(r"$" + f_label + r"$", fontsize=12)
    axs[0, 0].set_xlabel(r"$" + x_label + "$", fontsize=12)
    axs[0, 0].set_title(top_left_subtitle, fontsize=12)
    axs[0, 0].grid(True)

    # [Top-Right] f_ref(x) Vs x
    axs[0, 1].plot(x, f_ref, 'b-', lw=1.5)
    axs[0, 1].set_ylabel(r"$" + f_ref_label + r"$", fontsize=12)
    axs[0, 1].set_xlabel(r"$" + x_label + "$", fontsize=12)
    axs[0, 1].set_title(top_right_subtitle, fontsize=12)
    axs[0, 1].grid(True)

    # [Mid-Left] f(x) - f_ref(x) Vs x
    axs[1, 0].plot(x, f - f_ref, 'r-')
    axs[1, 0].plot(x, np.zeros(len(x)), 'k--', lw=0.5)
    axs[1, 0].set_ylabel(r"$" + f_label + r" - " + f_ref_label + r"$", fontsize=12)
    axs[1, 0].set_xlabel(r"$" + x_label + "$", fontsize=12)
    axs[1, 0].set_title("Differences", fontsize=12)
    axs[1, 0].grid(True)

    # [Mid-Right] (f(x) - f_ref(x)) / f_ref(x) Vs x
    f_ref_nonzero = np.empty_like(f_ref) * np.nan
    f_ref_nonzero_mask = np.abs(f_ref) > 1e-15  # f_ref != 0
    f_ref_nonzero[f_ref_nonzero_mask] = f_ref[f_ref_nonzero_mask]

    axs[1, 1].plot(x, ((f - f_ref) / f_ref_nonzero) * 100, 'r-', lw=1.5)
    axs[1, 1].plot(x, np.zeros(len(x)), 'k--', lw=0.5)
    axs[1, 1].set_ylabel(r"$ \frac{" + f_label + r" - " + f_ref_label + r"}{" + f_ref_label + r"}$ (%)", fontsize=12)
    axs[1, 1].set_xlabel(r"$" + x_label + "$", fontsize=12)
    axs[1, 1].set_title("Relative Differences", fontsize=12)
    axs[1, 1].grid(True)

    # [Bottom-Left] |f(x) - f_ref(x)| Vs x
    axs[2, 0].plot(x, np.abs(f - f_ref), 'r-')
    axs[2, 0].plot(x, np.zeros(len(x)), 'k--', lw=0.5)
    axs[2, 0].set_ylabel(r"$|" + f_label + r" - " + f_ref_label + r"|$", fontsize=12)
    axs[2, 0].set_xlabel(r"$" + x_label + "$", fontsize=12)
    axs[2, 0].set_title("Differences (absolute value)", fontsize=12)
    axs[2, 0].grid(True)

    # [Bottom-Right] |(f(x) - f_ref(x)) / f_ref(x)| Vs x
    axs[2, 1].plot(x, np.abs((f - f_ref) / f_ref_nonzero) * 100, 'r-', lw=1.5)
    axs[2, 1].plot(x, np.zeros(len(x)), 'k--', lw=0.5)
    axs[2, 1].set_ylabel(r"$ \left| \frac{" + f_label + r" - " + f_ref_label + r"}{" + f_ref_label + r"} \right|$ (%)",
                         fontsize=12)
    axs[2, 1].set_xlabel(r"$" + x_label + "$", fontsize=12)
    axs[2, 1].set_title("Relative Differences (absolute value)", fontsize=12)
    axs[2, 1].grid(True)

    # make the main title
    fig.suptitle(title, fontsize=15)

    # show the plot
    fig.tight_layout()
    fig.subplots_adjust(top=0.88)
    plt.show()


# -----------------------------------------------------------------------------#

def plot(x, f, **kwargs):
    """
    Basic plotting function a bit customized
    """

    # parsing optional parameters
    x_label = kwargs['x_label'] if 'x_label' in kwargs else r"$x$"
    f_label = kwargs['f_label'] if 'f_label' in kwargs else "f"
    title = kwargs['title'] if 'title' in kwargs else "f(x) Vs x"
    f_up = kwargs['f_up'] if 'f_up' in kwargs else None
    f_up_label = kwargs['f_up_label'] if 'f_up_label' in kwargs else 'f_up_label'
    f_down = kwargs['f_down'] if 'f_down' in kwargs else None
    f_down_label = kwargs['f_down_label'] if 'f_down_label' in kwargs else 'f_down_label'

    # define the figure
    fig, ax = plt.subplots(figsize=(10, 6))

    # f(x) Vs x
    ax.plot(x, f, 'b-', lw=1.5)
    ax.set_ylabel(f_label, fontsize=12)
    ax.set_xlabel(x_label, fontsize=12)
    ax.set_title(title, fontsize=15)
    ax.grid(True)

    if f_up is not None:
        ax.plot(x, f_up, 'g--', lw=0.5, label=f_up_label)

    if f_down is not None:
        ax.plot(x, f_down, 'r--', lw=0.5, label=f_down_label)

    # add legend
    if (f_up is not None) or (f_down is not None):
        ax.legend(loc='best', ncol=1)

    # show the plot
    fig.tight_layout()
    plt.show()



import setuptools

with open("README_PyPI.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyBlackScholesAnalytics",
    version="0.0.1",
    author="Gabriele Pompa",
    author_email="gabriele.pompa@gmail.com",
    description="Options and Option Strategies analytics under the Black-Scholes Model for educational purpose",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gabrielepompa88/pyBlackScholesAnalytics",
	project_urls={
	'Repository':  "https://github.com/gabrielepompa88/pyBlackScholesAnalytics"
	},
    packages=setuptools.find_packages(),
    classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Financial and Insurance Industry',
		'Intended Audience :: Education',
		'Intended Audience :: Science/Research',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
	keywords='finance black-scholes education',
    python_requires='>=3.6',
)


import unittest
import copy
import numpy as np
import pandas as pd
import pandas.testing as pd_test

from pyblackscholesanalytics.market.market import MarketEnvironment
from pyblackscholesanalytics.options.options import PlainVanillaOption
from pyblackscholesanalytics.portfolio.portfolio import Portfolio


class TestPortfolio(unittest.TestCase):
    """Class to test public methods of Portfolio class"""

    def setUp(self) -> None:

        # common market environment
        mkt_env = MarketEnvironment(t="01-06-2020")

        # underlying values to test
        S_vector = [60, 90, 120]

        # options maturities
        T_call = "31-12-2020"
        T_put = "30-06-2021"

        # time parameter
        t_range = pd.date_range(start=mkt_env.get_t(), end=T_call, periods=3)

        # pricing parameters
        self.params = {"S": S_vector, "t": t_range, "np_output": False}

        # options strikes
        K_put = 80
        K_call = 110

        # portfolio options positions
        self.call_pos = 2
        self.put_pos = -5

        # empty portfolio initialized
        self.ptf = Portfolio()

        # adding 2 long plain-vanilla call contracts
        self.call_opt = PlainVanillaOption(mkt_env, K=K_call, T=T_call)

        # adding 5 short plain-vanilla put contracts
        self.put_opt = PlainVanillaOption(mkt_env, option_type="put", K=K_put, T=T_put)

        # adding contracts to portfolio
        self.ptf.add_instrument(self.call_opt, self.call_pos)
        self.ptf.add_instrument(self.put_opt, self.put_pos)

    def test_portfolio_price(self):
        """Test that portfolio metrics matches algebraic sum of individual instruments' metrics.
        Metrics: price"""

        metrics = "price"

        test = getattr(self.ptf, metrics)(**self.params)
        expected = self.call_pos*getattr(self.call_opt, metrics)(**self.params) + \
                   self.put_pos*getattr(self.put_opt, metrics)(**self.params)

        pd_test.assert_frame_equal(test, expected)

    def test_portfolio_PnL(self):
        """Test that portfolio metrics matches algebraic sum of individual instruments' metrics.
        Metrics: PnL"""

        metrics = "PnL"

        test = getattr(self.ptf, metrics)(**self.params)
        expected = self.call_pos * getattr(self.call_opt, metrics)(**self.params) + \
                   self.put_pos * getattr(self.put_opt, metrics)(**self.params)

        pd_test.assert_frame_equal(test, expected)

    def test_portfolio_delta(self):
        """Test that portfolio metrics matches algebraic sum of individual instruments' metrics.
        Metrics: delta"""

        metrics = "delta"

        test = getattr(self.ptf, metrics)(**self.params)
        expected = self.call_pos * getattr(self.call_opt, metrics)(**self.params) + \
                   self.put_pos * getattr(self.put_opt, metrics)(**self.params)

        pd_test.assert_frame_equal(test, expected)

    def test_portfolio_gamma(self):
        """Test that portfolio metrics matches algebraic sum of individual instruments' metrics.
        Metrics: gamma"""

        metrics = "gamma"

        test = getattr(self.ptf, metrics)(**self.params)
        expected = self.call_pos * getattr(self.call_opt, metrics)(**self.params) + \
                   self.put_pos * getattr(self.put_opt, metrics)(**self.params)

        pd_test.assert_frame_equal(test, expected)

    def test_portfolio_vega(self):
        """Test that portfolio metrics matches algebraic sum of individual instruments' metrics.
        Metrics: vega"""

        metrics = "vega"

        test = getattr(self.ptf, metrics)(**self.params)
        expected = self.call_pos * getattr(self.call_opt, metrics)(**self.params) + \
                   self.put_pos * getattr(self.put_opt, metrics)(**self.params)

        pd_test.assert_frame_equal(test, expected)

    def test_portfolio_theta(self):
        """Test that portfolio metrics matches algebraic sum of individual instruments' metrics.
        Metrics: theta"""

        metrics = "theta"

        test = getattr(self.ptf, metrics)(**self.params)
        expected = self.call_pos * getattr(self.call_opt, metrics)(**self.params) + \
                   self.put_pos * getattr(self.put_opt, metrics)(**self.params)

        pd_test.assert_frame_equal(test, expected)

    def test_portfolio_rho(self):
        """Test that portfolio metrics matches algebraic sum of individual instruments' metrics.
        Metrics: rho"""

        metrics = "rho"

        test = getattr(self.ptf, metrics)(**self.params)
        expected = self.call_pos * getattr(self.call_opt, metrics)(**self.params) + \
                   self.put_pos * getattr(self.put_opt, metrics)(**self.params)

        pd_test.assert_frame_equal(test, expected)


if __name__ == '__main__':
    unittest.main()
